﻿using System;
using System.Collections.Generic;
using System.Data.OleDb;
using System.IO;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;

namespace WebDAVServer.Class1Listener
{
    class Folder : HierarchyItem, IFolder
    {
        public Folder(Request request, int id, string name, DateTime created, DateTime modified, string path)
            : base(request, id, name, created, modified, path.TrimEnd('/') + "/")
        {
        }

        #region IFolder Members

        public IHierarchyItem[] Children
        {
            get
            {
                List<IHierarchyItem> children = new List<IHierarchyItem>();

                OleDbConnection conn = new OleDbConnection(connStr);
                OleDbDataReader reader = null;
                conn.Open();

                try
                {
                    OleDbCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SELECT ID, ItemType, Name, Created, Modified"
                        + " FROM Repository"
                        + " WHERE Parent = ?";

                    cmd.Parameters.Add("@Parent", OleDbType.Integer).Value = ID;

                    reader = cmd.ExecuteReader();
                    while (reader.Read())
                    {
                        int childID = reader.GetInt32(reader.GetOrdinal("ID"));
                        ItemType childType = (ItemType)reader.GetByte(reader.GetOrdinal("ItemType"));
                        string childName = reader.GetString(reader.GetOrdinal("Name"));
                        DateTime childCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                        DateTime childModified = reader.GetDateTime(reader.GetOrdinal("Modified"));
                        switch (childType)
                        {
                            case ItemType.Resource:
                                children.Add(
                                    new Resource(request, childID, childName, childCreated, childModified, path + childName));
                                break;
                            case ItemType.Folder:
                                children.Add(
                                    new Folder(request, childID, childName, childCreated, childModified, path + childName + "/"));
                                break;
                        }
                    }
                }
                finally
                {
                    if (reader != null) reader.Close();
                    conn.Close();
                }

                return children.ToArray();
            }
        }

        public WebDAVResponse CreateFolder(string folderName)
        {
            CreateChild(folderName, ItemType.Folder);
            return new CreatedResponse();
        }

        public WebDAVResponse CreateResource(string resName)
        {
            CreateChild(resName, ItemType.Resource);
            return new CreatedResponse();
        }

        #endregion // IFolder Members

        #region IHierarchyItem Members

        public override WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
        {
            // copy this folder
            WebDAVResponse resp = base.CopyTo(folder, destName, deep);

            // copy children
            if (deep && resp.IsSuccess)
            {
                Folder destFolder = (Folder) folder;
                IFolder newDestFolder = destFolder.FindChild(destName) as IFolder;

                foreach (IHierarchyItem child in Children)
                {
                    child.CopyTo(newDestFolder, child.Name, deep);
                }
            }

            return resp;
        }

        public override WebDAVResponse Delete()
        {
            foreach (IHierarchyItem child in Children)
            {
                child.Delete();
            }

            return base.Delete();
        }

        public override WebDAVResponse MoveTo(IFolder folder, string destName)
        {
            Folder destFolder = (Folder) folder;
            IHierarchyItem destItem = destFolder.FindChild(destName);

            if (destItem != null)
            {
                if (destItem is IResource)
                {
                    destItem.Delete();
                    base.MoveTo(folder, destName);
                }
                else
                {
                    // copy this folder over existing destination folder
                    CopyTo(folder, destName, false);

                    IFolder newDestFolder = destItem as IFolder;

                    // copy children
                    foreach (IHierarchyItem child in Children)
                    {
                        child.MoveTo(newDestFolder, child.Name);
                    }

                    // delete this folder
                    Delete();
                }

                return new NoContentResponse();
            }
            else
            {
                base.MoveTo(folder, destName);
                return new CreatedResponse();
            }
        }

        #endregion

        #region Non-public Members

        internal IHierarchyItem FindChild(string childName)
        {
            HierarchyItem child = null;

            OleDbConnection conn = new OleDbConnection(connStr);
            OleDbDataReader reader = null;
            conn.Open();

            try
            {
                OleDbCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT ID, ItemType, Name, Created, Modified"
                    + " FROM Repository"
                    + " WHERE Parent = ?"
                    + " AND Name = ?";

                cmd.Parameters.Add("@Parent", OleDbType.Integer).Value = ID;
                cmd.Parameters.Add("@Name", OleDbType.WChar).Value = childName;

                reader = cmd.ExecuteReader();
                if (reader.Read())
                {
                    int itemID = reader.GetInt32(reader.GetOrdinal("ID"));
                    ItemType itemType = (ItemType)reader.GetByte(reader.GetOrdinal("ItemType"));
                    string itemName = reader.GetString(reader.GetOrdinal("Name"));
                    DateTime itemCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                    DateTime itemModified = reader.GetDateTime(reader.GetOrdinal("Modified"));

                    switch (itemType)
                    {
                        case ItemType.Resource:
                            child = new Resource(request, itemID, itemName, itemCreated, itemModified, path + itemName);
                            break;
                        case ItemType.Folder:
                            child = new Folder(request, itemID, itemName, itemCreated, itemModified, path + itemName + "/");
                            break;
                    }
                }
            }
            finally
            {
                if (reader != null) reader.Close();
                conn.Close();
            }

            return child;
        }

        private Resource CreateChild(string childName, ItemType itemType)
        {
            OleDbConnection conn = new OleDbConnection(connStr);
            conn.Open();

            try
            {
                OleDbCommand cmd = conn.CreateCommand();
                cmd.CommandText = "INSERT INTO Repository"
                    + " (Name, Created, Modified, Parent, ItemType)"
                    + " VALUES(?, ?, ?, ?, ?)";

                cmd.Parameters.Add("@Name", OleDbType.WChar).Value = childName;
                cmd.Parameters.Add("@Created", OleDbType.Date).Value = DateTime.UtcNow;
                cmd.Parameters.Add("@Modified", OleDbType.Date).Value = DateTime.UtcNow;
                cmd.Parameters.Add("@Parent", OleDbType.Integer).Value = ID;
                cmd.Parameters.Add("@ItemType", OleDbType.TinyInt).Value = itemType;

                cmd.ExecuteNonQuery();

                UpdateModified(conn);

                if (itemType == ItemType.Resource)
                {
                    cmd.CommandText = "SELECT @@IDENTITY";
                    cmd.Parameters.Clear();
                    OleDbDataReader reader = cmd.ExecuteReader();
                    int newID;
                    try
                    {
                        reader.Read();
                        try
                        {
                            newID = reader.GetInt32(0);
                        }
                        catch
                        {
                            newID = (int)reader.GetDecimal(0);
                        }
                    }
                    finally
                    {
                        reader.Close();
                    }
                    return new Resource(request, newID, childName, DateTime.UtcNow, DateTime.UtcNow, path + childName);
                }
                else
                {
                    return null;
                }
            }
            finally
            {
                conn.Close();
            }
        }

        #endregion // Non-public Members
    }
}
