﻿using System;
using System.Collections.Generic;
using System.Configuration;
using System.Data.OleDb;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;

namespace WebDAVServer.Class1Listener
{
    abstract class HierarchyItem : IHierarchyItem
    {
        protected Request request;
        protected int id;
        protected string name;
        protected DateTime created;
        protected DateTime modified;
        protected string path;
        protected string connStr = ConfigurationManager.ConnectionStrings["OleDbConnectionString"].ConnectionString;

        public HierarchyItem(Request request, int id, string name, DateTime created, DateTime modified, string path)
        {
            this.request = request;
            this.id = id;
            this.name = name;
            this.created = created;
            this.modified = modified;
            this.path = path;
        }

        #region IHierarchyItem Members

        public string Name
        {
            get { return name; }
        }

        public DateTime Created
        {
            get { return created; }
        }

        public DateTime Modified
        {
            get { return modified; }
        }

        public IFolder Parent
        {
            get
            {
                IFolder parent = null;

                OleDbConnection conn = new OleDbConnection(connStr);
                OleDbDataReader reader = null;
                conn.Open();

                try
                {
                    OleDbCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SELECT Parent FROM Repository WHERE ID = ?";
                    cmd.Parameters.Add("@ID", OleDbType.Integer).Value = ID;
                    int parentID = (int)cmd.ExecuteScalar();

                    if (parentID > 0)
                    {
                        cmd.CommandText = "SELECT Name, Created, Modified"
                            + " FROM Repository"
                            + " WHERE ID = ?";

                        cmd.Parameters["@ID"].Value = parentID;

                        reader = cmd.ExecuteReader();
                        if (reader.Read())
                        {
                            string parentName = reader.GetString(reader.GetOrdinal("Name"));
                            DateTime parentCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                            DateTime parentModified = reader.GetDateTime(reader.GetOrdinal("Modified"));

                            parent = new Folder(request, parentID, parentName, parentCreated, parentModified, path);
                        }
                    }
                }
                finally
                {
                    if (reader != null) reader.Close();
                    conn.Close();
                }

                return parent;
            }
        }

        public WebDAVResponse GetProperties(ref Property[] props)
        {
            OleDbConnection conn = new OleDbConnection(connStr);
            OleDbDataReader reader = null;
            conn.Open();

            try
            {
                OleDbCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT Name, Namespace, PropVal"
                    + " FROM Properties"
                    + " WHERE ItemID = ?";

                cmd.Parameters.Add("@ItemID", OleDbType.Integer).Value = ID;

                reader = cmd.ExecuteReader();
                if (props == null) // get all properties
                {
                    List<Property> l = new List<Property>();
                    while (reader.Read())
                    {
                        Property p = new Property();
                        p.Name = reader.GetString(reader.GetOrdinal("Name"));
                        p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                        p.Value = reader.GetString(reader.GetOrdinal("PropVal"));
                        l.Add(p);
                    }
                    props = l.ToArray();
                }
                else // get selected properties
                {
                    while (reader.Read())
                    {
                        Property p = new Property();
                        p.Name = reader.GetString(reader.GetOrdinal("Name"));
                        p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                        for (int i = 0; i < props.Length; i++)
                            if (p.Name == props[i].Name && p.Namespace == props[i].Namespace)
                            {
                                p.Value = reader.GetString(reader.GetOrdinal("PropVal"));
                                props[i] = p;
                                break;
                            }
                    }
                }
            }
            finally
            {
                if (reader != null) reader.Close();
                conn.Close();
            }

            return new OkResponse();
        }

        public WebDAVResponse GetPropertyNames(ref Property[] props)
        {
            OleDbConnection conn = new OleDbConnection(connStr);
            OleDbDataReader reader = null;
            conn.Open();

            try
            {
                OleDbCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT Name, Namespace"
                                  + " FROM Properties"
                                  + " WHERE ItemID = ?";

                cmd.Parameters.Add("@ItemID", OleDbType.Integer).Value = ID;
                reader = cmd.ExecuteReader();

                List<Property> l = new List<Property>();
                while (reader.Read())
                {
                    Property p = new Property();
                    p.Name = reader.GetString(reader.GetOrdinal("Name"));
                    p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                    l.Add(p);
                }
                props = l.ToArray();
            }
            finally
            {
                if (reader != null) reader.Close();
                conn.Close();
            }

            return new OkResponse();
        }

        public virtual string Path
        {
            get
            {
                return this.path;
            }
        }

        public WebDAVResponse UpdateProperties(Property[] setProps, Property[] delProps)
        {
            OleDbConnection conn = new OleDbConnection(connStr);
            conn.Open();

            MultipropResponse resp = new MultipropResponse();
            try
            {
                if (setProps != null)
                    foreach (Property p in setProps)
                    {
                        SetProperty(p, conn);
                        resp.AddResponses(new PropResponse(p, new OkResponse()));
                    }
                if (delProps != null)
                    foreach (Property p in delProps)
                    {
                        RemoveProperty(p.Name, p.Namespace, conn);
                        resp.AddResponses(new PropResponse(p, new OkResponse()));
                    }
                UpdateModified(conn);
            }
            finally
            {
                conn.Close();
            }

            return resp;
        }

        public virtual WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
        {
            WebDAVResponse resp;
            Folder destFolder = folder as Folder;

            if (destFolder == null)
                return new ConflictResponse();

            HierarchyItem destItem = destFolder.FindChild(destName) as HierarchyItem;

            OleDbConnection conn = new OleDbConnection(connStr);
            conn.Open();

            try
            {
                int destID;
                OleDbCommand cmd = conn.CreateCommand();

                if (destItem == null) // insert new item
                {
                    resp = new CreatedResponse();

                    // copy item
                    cmd.CommandText = "INSERT INTO Repository"
                        + " (Name, Created, Modified, Parent, ItemType, ContentType, Content)"
                        + " SELECT ?, ?, ?, ?, ItemType, ContentType, Content"
                        + " FROM Repository"
                        + " WHERE ID = ?";

                    cmd.Parameters.Add("@Name", OleDbType.WChar).Value = destName;
                    cmd.Parameters.Add("@Created", OleDbType.Date).Value = DateTime.UtcNow;
                    cmd.Parameters.Add("@Modified", OleDbType.Date).Value = DateTime.UtcNow;
                    cmd.Parameters.Add("@Parent", OleDbType.Integer).Value = destFolder.ID;
                    cmd.Parameters.Add("@ID", OleDbType.Integer).Value = ID;

                    cmd.ExecuteNonQuery();

                    cmd.CommandText = "SELECT @@IDENTITY";
                    cmd.Parameters.Clear();

                    OleDbDataReader reader = cmd.ExecuteReader();
                    try
                    {
                        reader.Read();
                        try
                        {
                            destID = reader.GetInt32(0);
                        }
                        catch
                        {
                            destID = (int)reader.GetDecimal(0);
                        }
                    }
                    finally
                    {
                        reader.Close();
                    }

                    destFolder.UpdateModified(conn);
                }
                else // update existing item
                {
                    resp = new NoContentResponse();

                    destID = destItem.ID;

                    // does not work with Access
                    cmd.CommandText = "UPDATE Repository SET"
                        + " Modified = ?"
                        + ", ItemType = src.ItemType"
                        + ", ContentType = src.ContentType"
                        + ", Content = src.Content"
                        + " FROM (SELECT * FROM Repository WHERE ID=?) src"
                        + " WHERE Repository.ID=?";

                    cmd.Parameters.Add("@Modified", OleDbType.Date).Value = DateTime.UtcNow;
                    cmd.Parameters.Add("@SrcID", OleDbType.Integer).Value = ID;
                    cmd.Parameters.Add("@DestID", OleDbType.Integer).Value = destID;

                    cmd.ExecuteNonQuery();

                    // remove old properties from the destination
                    cmd.CommandText = "DELETE FROM Properties WHERE ItemID = ?";

                    cmd.Parameters.Clear();
                    cmd.Parameters.Add("@ItemID", OleDbType.Integer).Value = destID;

                    cmd.ExecuteNonQuery();
                }

                // copy properties
                cmd.CommandText = "INSERT INTO Properties"
                    + " (ItemID, Name, Namespace, PropVal)"
                    + " SELECT ?, Name, Namespace, PropVal"
                    + " FROM Properties"
                    + " WHERE ItemID = ?";

                cmd.Parameters.Clear();
                cmd.Parameters.Add("@DestID", OleDbType.Integer).Value = destID;
                cmd.Parameters.Add("@SrcID", OleDbType.Integer).Value = ID;

                cmd.ExecuteNonQuery();
            }
            finally
            {
                conn.Close();
            }

            return resp;
        }

        public virtual WebDAVResponse Delete()
        {
            Folder parentFolder = Parent as Folder;

            OleDbConnection conn = new OleDbConnection(connStr);
            conn.Open();

            try
            {
                OleDbCommand cmd = conn.CreateCommand();

                cmd.Parameters.Add("@ItemID", OleDbType.Integer).Value = ID;

                cmd.CommandText = "DELETE FROM Properties WHERE ItemID = ?";
                cmd.ExecuteNonQuery();

                cmd.CommandText = "DELETE FROM Repository WHERE ID = ?";
                cmd.ExecuteNonQuery();

                if (parentFolder != null) parentFolder.UpdateModified(conn);
            }
            finally
            {
                conn.Close();
            }

            return new NoContentResponse();
        }

        public virtual WebDAVResponse MoveTo(IFolder folder, string destName)
        {
            Folder destFolder = (Folder) folder;
            Folder parentFolder = (Folder) Parent;

            OleDbConnection conn = new OleDbConnection(connStr);
            conn.Open();

            try
            {
                OleDbCommand cmd = conn.CreateCommand();

                cmd.CommandText = "UPDATE Repository SET"
                    + " Name = ?"
                    + ", Parent = ?"
                    + " WHERE ID = ?";

                cmd.Parameters.Add("@Name", OleDbType.WChar).Value = destName;
                cmd.Parameters.Add("@Parent", OleDbType.Integer).Value = destFolder.ID;
                cmd.Parameters.Add("@ID", OleDbType.Integer).Value = ID;

                cmd.ExecuteNonQuery();


                destFolder.UpdateModified(conn);
                parentFolder.UpdateModified(conn);

                name = destName;
            }
            finally
            {
                conn.Close();
            }

            // we don't know here whether to return CreatedResponse or NoContentResponse
            return null;
        }

        #endregion // IHierarchyItem Members

        #region Non-public Members

        private void SetProperty(Property prop, OleDbConnection conn)
        {
            OleDbCommand cmd = conn.CreateCommand();

            cmd.CommandText = "SELECT Count(*) FROM Properties"
                + " WHERE ItemID = ?"
                + " AND Name = ?"
                + " AND Namespace = ?";

            cmd.Parameters.Add("@ItemID", OleDbType.Integer).Value = ID;
            cmd.Parameters.Add("@Name", OleDbType.WChar).Value = prop.Name;
            cmd.Parameters.Add("@Namespace", OleDbType.WChar).Value = prop.Namespace;

            int count = (int)cmd.ExecuteScalar();

            if (count == 0) // insert
            {
                cmd.CommandText = "INSERT INTO Properties"
                    + " (ItemID, Name, Namespace, PropVal)"
                    + " VALUES(?, ?, ?, ?)";

                cmd.Parameters.Add("@PropVal", OleDbType.WChar).Value = prop.Value;
            }
            else // update
            {
                cmd.CommandText = "UPDATE Properties"
                    + " SET PropVal = ?"
                    + " WHERE ItemID = ?"
                    + " AND Name = ?"
                    + " AND Namespace = ?";

                cmd.Parameters.Clear();
                cmd.Parameters.Add("@PropVal", OleDbType.WChar).Value = prop.Value;
                cmd.Parameters.Add("@ItemID", OleDbType.Integer).Value = ID;
                cmd.Parameters.Add("@Name", OleDbType.WChar).Value = prop.Name;
                cmd.Parameters.Add("@Namespace", OleDbType.WChar).Value = prop.Namespace;
            }

            cmd.ExecuteNonQuery();
        }

        private void RemoveProperty(string propertyName, string ns, OleDbConnection conn)
        {
            OleDbCommand cmd = conn.CreateCommand();

            cmd.CommandText = "DELETE FROM Properties"
                + " WHERE ItemID = ?"
                + " AND Name = ?"
                + " AND Namespace = ?";

            cmd.Parameters.Add("@ItemID", OleDbType.Integer).Value = ID;
            cmd.Parameters.Add("@Name", OleDbType.WChar).Value = propertyName;
            cmd.Parameters.Add("@Namespace", OleDbType.WChar).Value = ns;

            cmd.ExecuteNonQuery();
        }

        internal int ID
        {
            get { return id; }
        }

        internal void UpdateModified(OleDbConnection conn)
        {
            OleDbCommand cmd = conn.CreateCommand();

            cmd.CommandText = "UPDATE Repository SET Modified = ? WHERE ID = ?";
            cmd.Parameters.Add("@Modified", OleDbType.Date).Value = DateTime.UtcNow;
            cmd.Parameters.Add("@ID", OleDbType.Integer).Value = ID;

            cmd.ExecuteNonQuery();
        }

        #endregion // Non-public Members
    }
}
