using System;
using System.Configuration;
using System.Net;
using System.ServiceProcess;
using System.Text;
using System.Threading;
using ITHit.WebDAV.Logger;

namespace WebDAVServer.DeltaV
{
    partial class DeltaVService : ServiceBase
    {
        private Thread thread;
        private bool listening = false;
        private static DigestAuthenticationProvider digestProvider = null;
        private static string authType = ConfigurationManager.AppSettings["AuthenticationType"];
        private static bool autoPutUnderVersionControl =
            Convert.ToBoolean(ConfigurationManager.AppSettings["AutoPutUnderVersionControl"]);

        #region Public Members

        public DeltaVService()
        {
            InitializeComponent();
        }

        public DeltaVService(bool isApplication)
        {
            listening = isApplication;
        }

        public void ThreadProc()
        {
            string uriPrefix = ConfigurationManager.AppSettings["ListenerPrefix"];
            HttpListener listener = new HttpListener();
            listener.Prefixes.Add(uriPrefix);

            switch (authType.ToLower())
            {
                case "none":
                    listener.AuthenticationSchemes = AuthenticationSchemes.Anonymous;
                    break;
                case "basic":
                    listener.AuthenticationSchemes = AuthenticationSchemes.Basic;
                    break;
                case "digest":
                    listener.AuthenticationSchemes = AuthenticationSchemes.Anonymous;
                    break;
                case "ntlm":
                    listener.AuthenticationSchemes = AuthenticationSchemes.Ntlm;
                    break;
                case "kerberos":
                    listener.AuthenticationSchemes = AuthenticationSchemes.IntegratedWindowsAuthentication;
                    break;
            }

            listener.Start();

            Console.WriteLine("Start listening on " + uriPrefix);
            Console.WriteLine("Press Control-C to stop listener...");

            while (listening)
            {
                IAsyncResult result = listener.BeginGetContext(new AsyncCallback(ListenerCallback), listener);
                result.AsyncWaitHandle.WaitOne();
            }
        }

        public static void ListenerCallback(IAsyncResult result)
        {
            HttpListener listener = (HttpListener)result.AsyncState;
            HttpListenerContext context = listener.EndGetContext(result);
            MacOsXPreprocessor.Process(context.Request); // fixes headers for Mac OS X v10.5.3 or later

            WDEngine engine = new WDEngine(context.User);
            context.Response.SendChunked = false;
            FileLogger.Level = LogLevel.Debug;
            
            // set custom handler to process GET requests to folders and display info about how to connect to server
            MyCustomGetHandler handler = new MyCustomGetHandler();
            handler.OriginalHandler = engine.RegisterMethodHandler("GET", handler);

            engine.AutoPutUnderVersionControl = autoPutUnderVersionControl;
            engine.IgnoreExceptions = false; // if any exception during the call to Engine.Run occurs exception will be thrown

            if (UserAutorized(context))
            {
                try
                {
                    engine.Run(context, listener.Prefixes);
                    engine.CommitTransaction();
                }
                catch
                {
                    engine.RollBackTransaction();
                }
                finally
                {
                    engine.CloseConnection();
                }
            }
            else
                context.Response.StatusCode = 401;

            if (context.Response.StatusCode == 401)
                ShowLoginDialog(context, context.Response);

            try
            {
                context.Response.Close();
            }
            catch
            {
                // client closed connection before the content was sent
            }
        }

        #endregion // Public Members

        #region ServiceBase Methods Override

        protected override void OnStart(string[] args)
        {
            listening = true;
            thread = new Thread(new ThreadStart(ThreadProc));
            thread.Start();
        }

        protected override void OnStop()
        {
            listening = false;
            thread.Abort();
            thread.Join();
        }

        #endregion // ServiceBase Methods Override

        #region Private Members

        private static void ShowLoginDialog(HttpListenerContext context, HttpListenerResponse response)
        {
            switch (authType.ToLower())
            {
                case "basic":
                    context.Response.AddHeader("WWW-Authenticate",
                                               "Basic Realm=\"My WebDAV Server\"");
                    break;
                case "digest":
                    digestProvider.SetChallenge(context.Response);
                    break;
                case "ntlm":
                    response.AddHeader("WWW-Authenticate", "NTLM");
                    break;
                case "kerberos":
                    response.AddHeader("WWW-Authenticate", "Negotiate");
                    break;
            }

            if (context.Request.HttpMethod.ToLower() != "options"
                && context.Request.HttpMethod.ToLower() != "head")
            {
                byte[] message = new UTF8Encoding().GetBytes("Access denied");
                context.Response.ContentLength64 = message.Length;
                context.Response.OutputStream.Write(message, 0, message.Length);
            }
        }

        private static bool UserAutorized(HttpListenerContext context)
        {
            switch (authType.ToLower())
            {
                case "basic":
                    HttpListenerBasicIdentity identity = (HttpListenerBasicIdentity)context.User.Identity;
                    string username = identity.Name;

                    // workaround for Windows Vista Basic authentication. User name may be submitted in the following format: Machine\User.
                    int ind = username.LastIndexOf('\\');
                    if (ind > 0)
                        username = username.Remove(0, ind + 1);

                    if ((username == "User1") && (identity.Password == "pwd"))
                        return true;
                    else
                        return false;
                case "digest":
                    if (context.Request.HttpMethod.ToLower() == "options")  // workaround for Windows Server 2003 and XP. Digest Authorization header is not attched to OPTIONS requests.
                        return true;
                    if (digestProvider == null)
                        digestProvider = new DigestAuthenticationProvider();
                    return digestProvider.RequestAuthenticated(context.Request) != null;
                default:
                    return true;
            }
        }

        #endregion // Private Members
    }
}
