using System;
using System.Collections.Generic;
using System.Net;
using System.Security.Cryptography;
using System.Security.Principal;
using System.Text;

namespace WebDAVServer.DeltaV
{
    public class DigestAuthenticationProvider
    {
        private bool isNonceStale = false;
        private static readonly string realm = "ITHitWebDAVServer";
        private static MD5 md5 = new MD5CryptoServiceProvider();
        private char[] trimSymbols = new char[] {' ', '\"'};

        public IPrincipal RequestAuthenticated(HttpListenerRequest request)
        {
            string authStr = request.Headers["Authorization"];

            if (authStr == null || authStr.Length == 0 || !authStr.ToLower().StartsWith("digest"))
                return null;

            authStr = authStr.Trim().Substring(7);

            Dictionary<string, string> reqInfo = new Dictionary<string, string>();

            string[] elems = authStr.Split(',');
            foreach (string elem in elems)
            {
                string[] parts = elem.Split(new char[] {'='}, 2);
                string key = parts[0].Trim(trimSymbols);
                string val = parts[1].Trim(trimSymbols);
                reqInfo.Add(key, val);
            }

            string clientUsername = (reqInfo.ContainsKey("username")) ? reqInfo["username"] : "";

            // workaround for Windows Vista Digest Authorization. User name may be submitted in the following format: Machine\\User.
            clientUsername = clientUsername.Replace("\\\\", "\\");
            reqInfo["username"] = clientUsername;
            string username = clientUsername;
            int ind = username.LastIndexOf('\\');
            if (ind > 0)
                username = username.Remove(0, ind + 1);
            
            string password;
            string[] roles;
            if (!GetPasswordAndRoles(username, out password, out roles))
                return null;

            string unhashedDigest = GenerateUnhashedDigest(password, reqInfo, request.HttpMethod);
            string hashedDigest = CreateMD5HashBinHex(unhashedDigest);

            isNonceStale = !IsNonceValid(reqInfo["nonce"]);

            if ((reqInfo["response"] != hashedDigest) || isNonceStale)
                return null;

            return new GenericPrincipal(new GenericIdentity(clientUsername, "digest"), roles);
        }

        #region Private Members

        private static string GenerateUnhashedDigest(string password, Dictionary<string, string> reqInfo, string httpMethod)
        {
            string A1 = String.Format("{0}:{1}:{2}", reqInfo["username"], realm, password);
            string HA1 = CreateMD5HashBinHex(A1);
            string A2 = String.Format("{0}:{1}", httpMethod, reqInfo["uri"]);
            string HA2 = CreateMD5HashBinHex(A2);

            string unhashedDigest;
            if (reqInfo["qop"] != null)
                unhashedDigest = String.Format("{0}:{1}:{2}:{3}:{4}:{5}",
                                               HA1,
                                               reqInfo["nonce"],
                                               reqInfo["nc"],
                                               reqInfo["cnonce"],
                                               reqInfo["qop"],
                                               HA2);
            else
                unhashedDigest = String.Format("{0}:{1}:{2}",
                                               HA1,
                                               reqInfo["nonce"],
                                               HA2);
            return unhashedDigest;
        }

        public void SetChallenge(HttpListenerResponse response)
        {
            if (response.StatusCode == 401)
            {
                string nonce = CreateNewNonce();

                StringBuilder stringBuilder = new StringBuilder("Digest");
                stringBuilder.Append(" realm=\"");
                stringBuilder.Append(realm);
                stringBuilder.Append("\", nonce=\"");
                stringBuilder.Append(nonce);
                stringBuilder.Append("\", opaque=\"0000000000000000\", stale=");
                stringBuilder.Append(isNonceStale ? "true" : "false");
                stringBuilder.Append(", algorithm=MD5, qop=\"auth\"");

                response.AddHeader("WWW-Authenticate", stringBuilder.ToString());
                response.StatusCode = 401;
            }
        }

        private static string CreateMD5HashBinHex(string val)
        {
            byte[] bHA1 = md5.ComputeHash(Encoding.ASCII.GetBytes(val));
            string HA1 = "";
            for (int i = 0; i < 16; i++)
                HA1 += String.Format("{0:x02}", bHA1[i]);
            return HA1;
        }

        private static string CreateNewNonce()
        {
            DateTime nonceTime = DateTime.Now + TimeSpan.FromMinutes(1);
            string expireStr = nonceTime.ToString("G");

            byte[] expireBytes = Encoding.ASCII.GetBytes(expireStr);
            string nonce = Convert.ToBase64String(expireBytes);

            nonce = nonce.TrimEnd(new Char[] { '=' });
            return nonce;
        }

        private static bool IsNonceValid(string nonce)
        {
            DateTime expireTime;

            int numPadChars = nonce.Length % 4;
            if (numPadChars > 0)
                numPadChars = 4 - numPadChars;
            string newNonce = nonce.PadRight(nonce.Length + numPadChars, '=');

            try
            {
                byte[] decodedBytes = Convert.FromBase64String(newNonce);
                string expireStr = Encoding.ASCII.GetString(decodedBytes);
                expireTime = DateTime.Parse(expireStr);
            }
            catch (FormatException)
            {
                return false;
            }

            return (DateTime.Now <= expireTime);
        }

        private static bool GetPasswordAndRoles(string username, out string password, out string[] roles)
        {
            roles = null;
            password = null;

            if (username == "User1")
            {
                password = "pwd";
                roles = new string[] { "admin" };
                return true;
            }

            return false;
        }

        #endregion // Private Members
    }
}
