using System;
using System.Collections;
using System.Data;
using System.Data.SqlClient;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;

namespace WebDAVServer.DeltaV
{
    public class Folder : HierarchyItem, IFolder, IFolderLock
    {
        public Folder(WDEngine engine, Guid id, string name, string path, DateTime created, DateTime modified)
            : base(engine, id, name, path, created, modified)
        {
        }

        #region IFolder Members

        public IHierarchyItem[] Children
        {
            get
            {
                ArrayList children = new ArrayList();

                SqlCommand cmd = fEngine.CreateNewCommand();
                cmd.CommandText = "SELECT ItemId, ItemType, Name, Created, Modified"
                                  + " FROM Item"
                                  + " WHERE ParentItemId = @Parent";
                cmd.Parameters.Add("@Parent", SqlDbType.UniqueIdentifier).Value = ItemId;

                using (SqlDataReader reader = cmd.ExecuteReader())
                {
                    while (reader.Read())
                    {
                        Guid childID = reader.GetGuid(reader.GetOrdinal("ItemId"));
                        ItemType childType = (ItemType) reader.GetInt32(reader.GetOrdinal("ItemType"));
                        string childName = reader.GetString(reader.GetOrdinal("Name"));
                        DateTime childCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                        DateTime childModified = reader.GetDateTime(reader.GetOrdinal("Modified"));

                        switch (childType)
                        {
                            case ItemType.Resource:
                                children.Add(
                                    new Resource(fEngine, childID, childName, fPath + childName, childCreated,
                                                 childModified));
                                break;
                            case ItemType.Folder:
                                children.Add(
                                    new Folder(fEngine, childID, childName, fPath + childName + "/", childCreated,
                                               childModified));
                                break;
                            case ItemType.LockNull:
                                children.Add(
                                    new LockNull(fEngine, childID, childName, fPath + childName, childCreated,
                                                 childModified));
                                break;
                        }
                    }
                }

                return (IHierarchyItem[]) children.ToArray(typeof (IHierarchyItem));
            }
        }

        public WebDAVResponse CreateResource(string name)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            CreateChild(name, ItemType.Resource);

            return new CreatedResponse();
        }

        public WebDAVResponse CreateFolder(string name)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            CreateChild(name, ItemType.Folder);

            return new CreatedResponse();
        }

        #endregion // IFolder Members

        #region IFolderLock Members

        public WebDAVResponse CreateLockNull(string newItemName, ref LockInfo lockInfo)
        {
            LockInfo[] locks = ActiveLocks;
            if (locks.Length > 0)
            {
                for (int i = 0; i < locks.Length; i++)
                    if (locks[i].Deep && (!lockInfo.Shared || !locks[i].Shared))
                        return new LockedResponse();

                if (!ClientHasToken)
                    return new LockedResponse();
            }
            HierarchyItem newItem;

            newItem = CreateChild(newItemName, ItemType.LockNull);
            newItem.Lock(ref lockInfo);

            return new CreatedResponse();
        }

        #endregion // IFolderLock Members

        #region IHierarchyItem Members

        public override WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
        {
            Folder destFolder = folder as Folder;
            if (destFolder == null)
                return new ConflictResponse();
            if (!destFolder.ClientHasToken)
                return new LockedResponse();

            IHierarchyItem destItem = destFolder.FindChild(destName);
            Folder newDestFolder;

            if (destItem != null)
            {
                if (DelWhenReplace && destItem is IResource)
                {
                    WebDAVResponse delResp = destItem.Delete();
                    if (!delResp.IsSuccess)
                        return delResp;
                    newDestFolder = CopyThisItem(destFolder, null, destName);
                }
                else
                {
                    newDestFolder = destItem as Folder;
                    if (newDestFolder == null)
                        return new ConflictResponse();
                }
            }
            else
            {
                newDestFolder = CopyThisItem(destFolder, null, destName);
            }
            // copy children
            MultistatusResponse mr = new MultistatusResponse();
            if (deep)
            {
                foreach (IHierarchyItem child in Children)
                {
                    HierarchyItem dbchild = child as HierarchyItem;
                    WebDAVResponse resp = dbchild.CopyTo(newDestFolder, child.Name, deep);
                    if (!resp.IsSuccess)
                        mr.AddResponses(new ItemResponse(dbchild.Path, resp));
                    else
                    {
                        MultistatusResponse mrchild = resp as MultistatusResponse;
                        if (mrchild != null)
                            mr.AddResponses(mrchild.Responses);
                    }
                }
            }

            if (mr.Responses.Length > 0)
                return mr;
            else if (destItem == null)
                return new CreatedResponse();
            else
                return new NoContentResponse();
        }

        public override WebDAVResponse MoveTo(IFolder folder, string destName)
        {
            Folder destFolder = folder as Folder;
            if (destFolder == null)
                return new ConflictResponse();

            Folder parent = Parent as Folder;
            if (parent == null)
                return new ConflictResponse();

            if (!ClientHasToken || !destFolder.ClientHasToken || !parent.ClientHasToken)
                return new LockedResponse();

            HierarchyItem destItem = destFolder.FindChild(destName);
            Folder newDestFolder;

            // copy this folder
            if (destItem != null)
            {
                if (DelWhenReplace && destItem is IResource)
                {
                    WebDAVResponse delResp = destItem.Delete();
                    if (!delResp.IsSuccess)
                        return delResp;
                    newDestFolder = CopyThisItem(destFolder, null, destName);
                }
                else
                {
                    newDestFolder = destItem as Folder;
                    if (newDestFolder == null)
                        return new ConflictResponse();
                }
            }
            else
            {
                newDestFolder = CopyThisItem(destFolder, null, destName);
            }
            // move children
            MultistatusResponse mr = new MultistatusResponse();
            foreach (IHierarchyItem child in Children)
            {
                HierarchyItem dbchild = child as HierarchyItem;
                WebDAVResponse resp = dbchild.MoveTo(newDestFolder, child.Name);
                if (!resp.IsSuccess)
                    mr.AddResponses(new ItemResponse(dbchild.Path, resp));
                else
                {
                    MultistatusResponse mrchild = resp as MultistatusResponse;
                    if (mrchild != null)
                        mr.AddResponses(mrchild.Responses);
                }
            }

            // delete this folder
            if (mr.Responses.Length > 0)
                return mr;
            else if (destItem == null)
            {
                DeleteThisItem(parent);
                return new CreatedResponse();
            }
            else
            {
                DeleteThisItem(parent);
                return new NoContentResponse();
            }
        }

        public override WebDAVResponse Delete()
        {
            Folder parent = Parent as Folder;
            if (parent == null)
                return new ConflictResponse();
            if (!parent.ClientHasToken)
                return new LockedResponse();

            if (!ClientHasToken)
                return new LockedResponse();

            MultistatusResponse mr = new MultistatusResponse();

            foreach (IHierarchyItem child in Children)
            {
                HierarchyItem dbchild = child as HierarchyItem;
                WebDAVResponse resp = dbchild.Delete();
                if (!resp.IsSuccess)
                    mr.AddResponses(new ItemResponse(dbchild.Path, resp));
                else
                {
                    MultistatusResponse mrchild = resp as MultistatusResponse;
                    if (mrchild != null)
                        mr.AddResponses(mrchild.Responses);
                }
            }

            if (mr.Responses.Length > 0)
                return mr;
            else
            {
                DeleteThisItem(parent);
                return new NoContentResponse();
            }
        }

        #endregion // IHierarchyItem Members

        #region Non-public Members

        private HierarchyItem CreateChild(string name, ItemType itemType)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();
            cmd.CommandText =
                "SET @Identity = NEWID()"
                + " INSERT INTO Item"
                + " (ItemId, Name, Created, Modified, ParentItemId, ItemType, Comment, CreatorDisplayName, AutoVersion, CheckinDuringUnlock, TotalContentLength, CheckinOnFileComplete)"
                + " VALUES(@Identity, @Name, GETUTCDATE(), GETUTCDATE(), @Parent, @ItemType, '', @CreatorDisplayName, 0, 0, 0, 0);";

            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = name;
            cmd.Parameters.Add("@Parent", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.Parameters.Add("@ItemType", SqlDbType.Int).Value = itemType;
            cmd.Parameters.Add("@CreatorDisplayName", SqlDbType.NVarChar).Value = CurrentUserName;
            SqlParameter idParm = cmd.Parameters.Add("@Identity", SqlDbType.UniqueIdentifier);
            idParm.Direction = ParameterDirection.Output;

            cmd.ExecuteNonQuery();
           
            Guid newID = (Guid) idParm.Value;

            //UpdateModified(); do not update time for folder as transaction will block concurrent files upload

            switch (itemType)
            {
                case ItemType.Resource:
                    return new Resource(fEngine, newID, name, fPath + name, DateTime.UtcNow, DateTime.UtcNow);
                case ItemType.Folder:
                    // do not need to return created folder
                    return null;
                case ItemType.LockNull:
                    return new LockNull(fEngine, newID, name, fPath + name, DateTime.UtcNow, DateTime.UtcNow);
                default:
                    return null;
            }
        }

        internal HierarchyItem FindChild(string childName)
        {
            HierarchyItem hierarchyItem = null;

            SqlCommand cmd = fEngine.CreateNewCommand();
            cmd.CommandText = "SELECT ItemId, ItemType, Name, Created, Modified"
                              + " FROM Item"
                              + " WHERE ParentItemId = @Parent"
                              + " AND Name = @Name";
            cmd.Parameters.Add("@Parent", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = childName;

            using (SqlDataReader reader = cmd.ExecuteReader())
            {
                if (reader.Read())
                {
                    Guid itemID = reader.GetGuid(reader.GetOrdinal("ItemId"));
                    ItemType itemType = (ItemType) reader.GetInt32(reader.GetOrdinal("ItemType"));
                    string itemName = reader.GetString(reader.GetOrdinal("Name"));
                    DateTime itemCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                    DateTime itemModified = reader.GetDateTime(reader.GetOrdinal("Modified"));

                    switch (itemType)
                    {
                        case ItemType.Resource:
                            hierarchyItem =
                                new Resource(fEngine, itemID, itemName, fPath + itemName,
                                             itemCreated, itemModified);
                            break;
                        case ItemType.Folder:
                            hierarchyItem =
                                new Folder(fEngine, itemID, itemName, fPath + itemName + "/",
                                           itemCreated, itemModified);
                            break;
                        case ItemType.LockNull:
                            hierarchyItem =
                                new LockNull(fEngine, itemID, itemName, fPath + itemName,
                                             itemCreated, itemModified);
                            break;
                    }
                }
            }

            return hierarchyItem;
        }

        internal void RemoveTree()
        {
            foreach (IHierarchyItem child in Children)
            {
                Folder childFolder = child as Folder;
                if (childFolder != null)
                    childFolder.RemoveTree();
                else
                    (child as HierarchyItem).DeleteThisItem();
            }
            DeleteThisItem();
        }

        internal bool ClientHasTokenForTree()
        {
            if (!ClientHasToken)
                return false;
            foreach (IHierarchyItem child in Children)
            {
                Folder childFolder = child as Folder;
                if (childFolder != null)
                {
                    if (!childFolder.ClientHasTokenForTree())
                        return false;
                }
                else
                {
                    if (!(child as HierarchyItem).ClientHasToken)
                        return false;
                }
            }
            return true;
        }

        #endregion // Non-public Members
    }
}