using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using System.IO;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;

namespace WebDAVServer.DeltaV
{
    public abstract class HierarchyItem : IHierarchyItem, ILock
    {
        protected WDEngine fEngine;
        protected Guid fItemId;
        protected string fName;
        protected string fPath;
        protected DateTime fCreated;
        protected DateTime fModified;
        protected static string fRepositoryPath = ConfigurationManager.AppSettings["RepositoryPath"].TrimEnd('\\') + "\\";

        public HierarchyItem(WDEngine engine, Guid id, string name, string path,
            DateTime created, DateTime modified)
        {
            this.fEngine = engine;
            this.fItemId = id;
            this.fName = name;
            this.fPath = path;
            this.fCreated = created;
            this.fModified = modified;
        }

        #region IHierarchyItem Members

        public string Name
        {
            get { return fName; }
        }

        public DateTime Created
        {
            get { return fCreated; }
        }

        public DateTime Modified
        {
            get { return fModified; }
        }

        public IFolder Parent
        {
            get
            {
                IFolder parent = null;

                SqlCommand cmd = fEngine.CreateNewCommand();
                cmd.CommandText = "SELECT ParentItemId FROM Item WHERE ItemId = @ItemId";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = ItemId;
                Guid parentID = (Guid) cmd.ExecuteScalar();

                if (parentID != Guid.Empty)
                {
                    cmd.CommandText = "SELECT Name, Created, Modified"
                                      + " FROM Item"
                                      + " WHERE ItemId = @ItemId";

                    cmd.Parameters["@ItemId"].Value = parentID;

                    using (SqlDataReader reader = cmd.ExecuteReader())
                    {
                        if (reader.Read())
                        {
                            string parentName = reader.GetString(reader.GetOrdinal("Name"));
                            DateTime parentCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                            DateTime parentModified = reader.GetDateTime(reader.GetOrdinal("Modified"));

                            int ln = fName.Length;
                            int lp = fPath.Length;
                            parent =
                                new Folder(fEngine, parentID, parentName, fPath.Remove(lp - ln, ln), parentCreated,
                                           parentModified);
                        }
                    }
                }

                return parent;
            }
        }

        public abstract WebDAVResponse CopyTo(IFolder folder, string destName, bool deep);

        public abstract WebDAVResponse MoveTo(IFolder folder, string destName);

        public abstract WebDAVResponse Delete();

        public WebDAVResponse GetProperties(ref Property[] props)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();
            cmd.CommandText = "SELECT Name, Namespace, PropVal"
                              + " FROM Property"
                              + " WHERE ItemID = @ItemID";

            cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = ItemId;

            using (SqlDataReader reader = cmd.ExecuteReader())
            {
                if (props == null) // get all properties
                {
                    List<Property> l = new List<Property>();
                    while (reader.Read())
                    {
                        Property p = new Property();
                        p.Name = reader.GetString(reader.GetOrdinal("Name"));
                        p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                        p.Value = reader.GetString(reader.GetOrdinal("PropVal"));
                        l.Add(p);
                    }
                    props = l.ToArray();
                }
                else // get selected properties
                {
                    while (reader.Read())
                    {
                        Property p = new Property();
                        p.Name = reader.GetString(reader.GetOrdinal("Name"));
                        p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                        for (int i = 0; i < props.Length; i++)
                            if (p.Name == props[i].Name && p.Namespace == props[i].Namespace)
                            {
                                p.Value = reader.GetString(reader.GetOrdinal("PropVal"));
                                props[i] = p;
                                break;
                            }
                    }
                }
            }

            return new OkResponse();
        }

        public virtual WebDAVResponse UpdateProperties(Property[] setProps, Property[] delProps)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            MultipropResponse resp = new MultipropResponse();

            if (setProps != null)
                foreach (Property p in setProps)
                {
                    SetProperty(p); // create or update property
                    resp.AddResponses(new PropResponse(p, new OkResponse()));
                }

            if (delProps != null)
                foreach (Property p in delProps)
                {
                    RemoveProperty(p.Name, p.Namespace);
                    resp.AddResponses(new PropResponse(p, new OkResponse()));
                }

            UpdateModified();

            return resp;
        }

        public virtual string Path
        {
             get
             {
                 return this.fPath;
             }
        }

        public WebDAVResponse GetPropertyNames(ref Property[] props)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();
            cmd.CommandText = "SELECT Name, Namespace"
                              + " FROM Property"
                              + " WHERE ItemID = @ItemID";

            cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = ItemId;

            using (SqlDataReader reader = cmd.ExecuteReader())
            {
                List<Property> l = new List<Property>();
                while (reader.Read())
                {
                    Property p = new Property();
                    p.Name = reader.GetString(reader.GetOrdinal("Name"));
                    p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                    l.Add(p);
                }
                props = l.ToArray();
            }

            return new OkResponse();
        }

        #endregion // IHierarchyItem Members

        #region ILock Members

        public WebDAVResponse Lock(ref LockInfo lockInfo)
        {
            if (ItemHasLock(lockInfo.Shared))
                return new LockedResponse();

            if (lockInfo.Deep)
            {
                // check if no items are locked in this subtree
                MultistatusResponse mr = FindLocksDown(this, lockInfo.Shared);
                if (mr.Responses.Length > 0)
                    return mr;
            }

            SqlCommand cmd = fEngine.CreateNewCommand();
            cmd.CommandText = "INSERT INTO Lock (ItemID,Token,Shared,Deep,Expires,Owner)"
                              + " VALUES(@ItemID, @Token, @Shared, @Deep, @Expires, @Owner)";

            lockInfo.Token = Guid.NewGuid().ToString();

            cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.Parameters.Add("@Token", SqlDbType.Char).Value = lockInfo.Token;
            cmd.Parameters.Add("@Shared", SqlDbType.Bit).Value = lockInfo.Shared;
            cmd.Parameters.Add("@Deep", SqlDbType.Bit).Value = lockInfo.Deep;
            cmd.Parameters.Add("@Expires", SqlDbType.DateTime);
            if (lockInfo.Timeout >= 0)
            {
                cmd.Parameters["@Expires"].Value = DateTime.Now.AddSeconds(lockInfo.Timeout);
            }
            else
            {
                cmd.Parameters["@Expires"].Value = SqlDateTime.Null;
            }
            cmd.Parameters.Add("@Owner", SqlDbType.NVarChar).Value = lockInfo.Owner;

            cmd.ExecuteNonQuery();

            return new OkResponse();
        }

        public WebDAVResponse RefreshLock(ref LockInfo lockInfo)
        {
            LockInfo[] locks = ActiveLocks;

            int lockIndex;
            for (lockIndex = 0; lockIndex < locks.Length; lockIndex++)
                if (lockInfo.Token == locks[lockIndex].Token) break;

            if (lockIndex == locks.Length)
                return new PreconditionFailedResponse();

            SqlCommand cmd = fEngine.CreateNewCommand();
            cmd.CommandText = "UPDATE Lock SET Expires = @Expires WHERE Token = @Token";
            cmd.Parameters.Add("@Expires", SqlDbType.DateTime);
            
            if (lockInfo.Timeout >= 0)
                cmd.Parameters["@Expires"].Value = DateTime.Now.AddSeconds(lockInfo.Timeout);
            else
                cmd.Parameters["@Expires"].Value = SqlBoolean.Null;

            cmd.Parameters.Add("@Token", SqlDbType.Char).Value = lockInfo.Token;
            cmd.ExecuteNonQuery();

            // return info about lock to WebDAV engine
            lockInfo.Shared = locks[lockIndex].Shared;
            lockInfo.Deep = locks[lockIndex].Deep;
            lockInfo.Owner = locks[lockIndex].Owner;

            return new OkResponse();
        }

        public WebDAVResponse Unlock(string lockToken)
        {
            LockInfo[] locks = ActiveLocks;
            int i;

            for (i = 0; i < locks.Length; i++)
                if (locks[i].Token == lockToken) break;

            if (i == locks.Length)
                return new PreconditionFailedResponse();

            if (this is ILockNull)
            {
                // delete lock-null item
                DeleteThisItem();
            }
            else
            {
                // remove lock from existing item
                SqlCommand cmd = fEngine.CreateNewCommand();
                cmd.CommandText = "DELETE FROM Lock WHERE Token = @Token";
                cmd.Parameters.Add("@Token", SqlDbType.NVarChar).Value = lockToken;
                cmd.ExecuteNonQuery();
            }

            return new NoContentResponse();
        }

        public LockInfo[] ActiveLocks
        {
            get
            {
                Guid itemID = ItemId;
                ArrayList l = new ArrayList();

                SqlCommand cmd = fEngine.CreateNewCommand();
                cmd.CommandText = "SELECT ParentItemId FROM Item WHERE ItemId = @ItemId";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier);
                cmd.Prepare();

                l.AddRange(GetLocks(itemID, false)); // get all locks
                while (true)
                {
                    cmd.Parameters["@ItemId"].Value = itemID;
                    itemID = (Guid) cmd.ExecuteScalar();
                    if (itemID == Guid.Empty) break;
                    l.AddRange(GetLocks(itemID, true)); // get only deep locks
                }

                return (LockInfo[]) l.ToArray(typeof (LockInfo));
            }
        }

        #endregion // ILock Members

        #region user properties methods

        protected void SetProperty(Property prop)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();
            
            cmd.CommandText = "SELECT Count(*) FROM Property"
                              + " WHERE ItemID = @ItemID"
                              + " AND Name = @Name"
                              + " AND Namespace = @Namespace";

            cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = prop.Name;
            cmd.Parameters.Add("@Namespace", SqlDbType.NVarChar).Value = prop.Namespace;

            int count = (int) cmd.ExecuteScalar();

            if (count == 0) // insert
            {
                cmd.CommandText = "INSERT INTO Property"
                                  + " (ItemID, Name, Namespace, PropVal)"
                                  + " VALUES(@ItemID, @Name, @Namespace, @PropVal)";

                cmd.Parameters.Add("@PropVal", SqlDbType.NVarChar).Value = prop.Value;
            }
            else // update
            {
                cmd.CommandText = "UPDATE Property"
                                  + " SET PropVal = @PropVal"
                                  + " WHERE ItemID = @ItemID"
                                  + " AND Name = @Name"
                                  + " AND Namespace = @Namespace";

                cmd.Parameters.Add("@PropVal", SqlDbType.NVarChar).Value = prop.Value;
            }

            cmd.ExecuteNonQuery();
        }

        protected void RemoveProperty(string name, string ns)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();
            
            cmd.CommandText = "DELETE FROM Property"
                              + " WHERE ItemID = @ItemID"
                              + " AND Name = @Name"
                              + " AND Namespace = @Namespace";

            cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = name;
            cmd.Parameters.Add("@Namespace", SqlDbType.NVarChar).Value = ns;

            cmd.ExecuteNonQuery();
        }

        #endregion

        #region copy/move methods

        internal Folder CopyThisItem(Folder destFolder, HierarchyItem destItem, string destName)
        {
            // returns created folder, if any, otherwise null
            Folder createdFolder = null;

            SqlCommand cmd = fEngine.CreateNewCommand();

            Guid destID;
            if (destItem == null) // insert new item
            {
                // copy item
                cmd.CommandText = "SET @Identity = NEWID()"
                                  + "INSERT INTO Item"
                                  +
                                  " (ItemId, Name, Created, Modified, ParentItemId, ItemType, ContentType, CreatorDisplayName, Comment, AutoVersion, CheckinDuringUnlock, TotalContentLength, CheckinOnFileComplete)"
                                  +
                                  " SELECT @Identity, @Name, GETUTCDATE(), GETUTCDATE(), @Parent, ItemType, ContentType, CreatorDisplayName, Comment, AutoVersion, CheckinDuringUnlock, TotalContentLength, CheckinOnFileComplete"
                                  + " FROM Item"
                                  + " WHERE ItemId = @ItemId;";

                cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = destName;
                cmd.Parameters.Add("@Parent", SqlDbType.UniqueIdentifier).Value = destFolder.ItemId;
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = ItemId;
                SqlParameter idParm = cmd.Parameters.Add("@Identity", SqlDbType.UniqueIdentifier);
                idParm.Direction = ParameterDirection.Output;

                cmd.ExecuteNonQuery();

                destID = (Guid) idParm.Value;

                destFolder.UpdateModified();

                if (this is IFolder)
                {
                    createdFolder = new Folder(fEngine, destID, destName, destFolder.fPath + destName + "/",
                                               DateTime.UtcNow, DateTime.UtcNow);
                }
                else
                {
                    Directory.CreateDirectory(fRepositoryPath + destID);
                }
            }
            else // update existing destination
            {
                destID = destItem.ItemId;

                cmd.CommandText = "UPDATE Item SET"
                                  + " Modified = GETUTCDATE()"
                                  + ", ItemType = src.ItemType"
                                  + ", ContentType = src.ContentType"
                                  + " FROM (SELECT * FROM Item WHERE ItemId=@SrcID) src"
                                  + " WHERE Item.ItemId=@DestID";

                cmd.Parameters.Add("@SrcID", SqlDbType.UniqueIdentifier).Value = ItemId;
                cmd.Parameters.Add("@DestID", SqlDbType.UniqueIdentifier).Value = destID;

                cmd.ExecuteNonQuery();

                // remove old properties from the destination
                cmd.CommandText = "DELETE FROM Property WHERE ItemID = @ItemID";

                cmd.Parameters.Clear();
                cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = destID;

                cmd.ExecuteNonQuery();
            }

            if (this is IResource)
            {
                string destPath = fRepositoryPath + destID + "\\current";
                if (File.Exists(destPath))
                    File.Delete(destPath);
                File.Copy(fRepositoryPath + ItemId + "\\current", destPath);
            }

            // copy properties
            cmd.CommandText = "INSERT INTO Property"
                              + " (ItemID, Name, Namespace, PropVal)"
                              + " SELECT @DestID, Name, Namespace, PropVal"
                              + " FROM Property"
                              + " WHERE ItemID = @SrcID";

            cmd.Parameters.Clear();
            cmd.Parameters.Add("@SrcID", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.Parameters.Add("@DestID", SqlDbType.UniqueIdentifier).Value = destID;
            cmd.ExecuteNonQuery();

            return createdFolder;
        }

        internal void MoveThisItem(Folder destFolder, string destName, Folder parent)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();
            // move item
            cmd.CommandText = "UPDATE Item SET"
                              + " Name = @Name"
                              + ", ParentItemId = @Parent"
                              + " WHERE ItemId = @ItemID";

            cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = destName;
            cmd.Parameters.Add("@Parent", SqlDbType.UniqueIdentifier).Value = destFolder.ItemId;

            cmd.ExecuteNonQuery();

            parent.UpdateModified();
            destFolder.UpdateModified();
            fName = destName;
        }

        #endregion // copy/move methods

        #region delete methods

        internal void DeleteThisItem()
        {
            this.DeleteThisItem(Parent as Folder);
        }

        internal void DeleteThisItem(Folder parentFolder)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();

            cmd.CommandText = "DELETE FROM Item WHERE ItemId = @ItemID";
            cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.ExecuteNonQuery();

            if (parentFolder != null) parentFolder.UpdateModified();

            Resource res = this as Resource;
            if (res != null) res.RemoveFromDisk();
        }

        #endregion

        #region lock methods

        private ArrayList GetLocks(Guid itemId, bool onlyDeep)
        {
            ArrayList l = new ArrayList(); // list of active locks
            ArrayList expl = new ArrayList(); // list of expired locks

            SqlCommand cmd = fEngine.CreateNewCommand();
            if (onlyDeep)
            {
                cmd.CommandText = "SELECT Token, Shared, Deep, Expires, Owner"
                                  + " FROM Lock"
                                  + " WHERE ItemID = @ItemID"
                                  + " AND Deep = @Deep";
                cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = itemId;
                cmd.Parameters.Add("@Deep", SqlDbType.Bit).Value = true;
            }
            else
            {
                cmd.CommandText = "SELECT Token, Shared, Deep, Expires, Owner FROM Lock WHERE ItemID = @ItemID";
                cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = itemId;
            }

            using (SqlDataReader reader = cmd.ExecuteReader())
            {
                while (reader.Read())
                {
                    LockInfo li = new LockInfo();
                    li.Token = reader.GetString(reader.GetOrdinal("Token"));
                    li.Shared = reader.GetBoolean(reader.GetOrdinal("Shared"));
                    li.Deep = reader.GetBoolean(reader.GetOrdinal("Deep"));
                    if (reader.IsDBNull(reader.GetOrdinal("Expires")))
                    {
                        li.Timeout = -1;
                    }
                    else
                    {
                        DateTime expires = reader.GetDateTime(reader.GetOrdinal("Expires"));
                        if (expires <= DateTime.Now)
                        {
                            li.Timeout = 0;
                        }
                        else
                            li.Timeout = (expires.Ticks - DateTime.Now.Ticks)/10000000;
                    }
                    li.Owner = reader.GetString(reader.GetOrdinal("Owner"));

                    if (li.Timeout == 0)
                        expl.Add(li);
                    else
                        l.Add(li);
                }
            }

            //if (expl.Count > 0) // unlock all expired locks
            //    foreach (LockInfo li in expl)
            //        Unlock(li.Token);

            return l;
        }

        internal bool ClientHasToken
        {
            get
            {
                LockInfo[] itemLocks = ActiveLocks;
                if (itemLocks.Length == 0)
                    return true;
                
                StringCollection clientLockTokens = fEngine.Request.ClientLockTokens;
                for (int i = 0; i < clientLockTokens.Count; i++)
                    for (int j = 0; j < itemLocks.Length; j++)
                        if (clientLockTokens[i] == itemLocks[j].Token)
                            return true;
                return false;
            }
        }

        protected bool ItemHasLock(bool skipShared)
        {
            LockInfo[] locks = ActiveLocks;
            if (locks.Length == 0)
                return false;
            bool allShared = true;
            foreach (LockInfo lockInfo in locks)
                if (!lockInfo.Shared)
                {
                    allShared = false;
                    break;
                }
            return !(skipShared && allShared);
        }

        protected static MultistatusResponse FindLocksDown(IHierarchyItem root, bool skipShared)
        {
            MultistatusResponse mr = new MultistatusResponse();
            IFolder folder = root as IFolder;
            if (folder != null)
                foreach (IHierarchyItem child in folder.Children)
                {
                    HierarchyItem dbchild = child as HierarchyItem;
                    if (dbchild.ItemHasLock(skipShared))
                        mr.AddResponses(new ItemResponse(dbchild.Path, new LockedResponse()));
                    MultistatusResponse mrchild = FindLocksDown(child, skipShared);
                    mr.AddResponses(mrchild.Responses);
                }
            return mr;
        }

        #endregion

        #region Internal Members

        internal void UpdateModified()
        {
            SqlCommand cmd = fEngine.CreateNewCommand();

            cmd.CommandText = "UPDATE Item SET Modified = GETUTCDATE() WHERE ItemId = @ItemId";
            cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = ItemId;

            cmd.ExecuteNonQuery();
        }

        internal Guid ItemId
        {
            get { return fItemId; }
        }

        #endregion // Internal Members

        #region Protected Members

        protected static bool DelWhenReplace
        {
            get
            {
                switch (ConfigurationManager.AppSettings["DeleteWhenReplace"])
                {
                    case "true":
                        return true;
                    case "false":
                        return false;
                    default:
                        return true;
                }
            }
        }

        protected string CurrentUserName
        {
            get
            {
                string name = string.Empty;
                
                if (fEngine.User != null && fEngine.User.Identity != null)
                        name = fEngine.User.Identity.Name;
                return name;
            }
        }

        #endregion // Protected Members

    }
}