using System;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Diagnostics;
using System.IO;
using System.Net;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.DeltaV;
using ITHit.WebDAV.Server.Response;
using ITHit.WebDAV.Server.ResumableUpload;

namespace WebDAVServer.DeltaV
{
    public class Resource : HierarchyItem, IResource, IVersionableItem, IResumableUpload, IUploadProgress
    {
        private string fFilePath;
        private string fItemDir;

        private static AutoVersion fAutoVersionMode =
            (AutoVersion) Enum.Parse(typeof (AutoVersion),
                                     ConfigurationManager.AppSettings["AutoVersionMode"]);

        public Resource(WDEngine engine, Guid ID, string name, string path, DateTime created, DateTime modified)
            : base(engine, ID, name, path, created, modified)
        {
            fItemDir = fRepositoryPath + fItemId;
            fFilePath = fItemDir + "\\current";
        }

        #region IResource Members

        public string ContentType
        {
            get
            {
                string contentType = GetDbField<string>("ContentType", null);

                if (string.IsNullOrEmpty(contentType))
                    contentType =
                        MimeType.GetMimeType(System.IO.Path.GetExtension(this.fName))
                        ?? "application/octet-stream";
                return contentType;
            }
        }

        public long ContentLength
        {
            get
            {
                if (File.Exists(fFilePath))
                {
                    FileInfo file = new FileInfo(fFilePath);
                    return file.Length;
                }
                else
                {
                    return 0;
                }
            }
        }

        public WebDAVResponse WriteToStream(Stream output, long byteStart, long count)
        {
            long bufSize = 1048576; // 1Mb

            FileInfo file = new FileInfo(fFilePath);
            byte[] buffer = new byte[bufSize];

            using (FileStream fileStream 
                = file.Open(FileMode.Open, FileAccess.Read, FileShare.Read))
            {
                fileStream.Seek(byteStart, SeekOrigin.Begin);
                int bytesRead;
                while ((bytesRead = fileStream.Read(buffer, 0, (int) (count > bufSize ? bufSize : count))) > 0)
                {
                    try
                    {
                        output.Write(buffer, 0, bytesRead);
                    }
                    catch (HttpListenerException ex)
                    {
                        if ((ex.ErrorCode == 1229) || (ex.ErrorCode == 64))
                        {
                            // client closed connection
                            // 1. ErrorCode=1229. An operation was attempted on a nonexistent network connection.
                            // 2. ErrorCode=64. The specified network name is no longer available.
                            return new OkResponse();
                        }
                        throw;
                    }
                    count -= bytesRead;
                }
            }

            return new OkResponse();
        }

        public WebDAVResponse SaveFromStream(Stream content, string contentType)
        {
            //This won't be ever called as soon as we implement IResumableUpload.            
            throw new NotImplementedException("We implement IResumableUpload so this method should not be called.");        
        }

        #endregion // IResource Members

        #region IHierarchyItem Members

        public override WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
        {
            Folder destFolder = folder as Folder;
            if (destFolder == null)
                return new ConflictResponse();
            if (!destFolder.ClientHasToken)
                return new LockedResponse();

            HierarchyItem destItem = destFolder.FindChild(destName);
            if (destItem != null)
            {
                if (DelWhenReplace && destItem is IFolder)
                {
                    Folder destItemF = destItem as Folder;
                    if (!destItemF.ClientHasTokenForTree())
                        return new LockedResponse();
                    destItemF.RemoveTree();
                    CopyThisItem(destFolder, null, destName);
                    return new NoContentResponse();
                }
                else if (destItem is IResource)
                {
                    if (destItem.ClientHasToken)
                    {
                        CopyThisItem(null, destItem, destName);
                        return new NoContentResponse();
                    }
                    else // destination is locked
                        return new LockedResponse();
                }
                else
                    return new ConflictResponse();
            }
            else
            {
                CopyThisItem(destFolder, null, destName);
                return new CreatedResponse();
            }
        }

        public override WebDAVResponse Delete()
        {
            Folder parent = Parent as Folder;
            if (parent == null)
                return new ConflictResponse();
            if (!parent.ClientHasToken)
                return new LockedResponse();

            if (!ClientHasToken)
                return new LockedResponse();

            DeleteThisItem(parent);
            return new NoContentResponse();
        }

        public override WebDAVResponse MoveTo(IFolder folder, string destName)
        {
            Folder destFolder = folder as Folder;
            if (destFolder == null)
                return new ConflictResponse();

            Folder parent = Parent as Folder;
            if (parent == null)
                return new ConflictResponse();

            if (!ClientHasToken || !destFolder.ClientHasToken || !parent.ClientHasToken)
                return new LockedResponse();

            HierarchyItem destItem = destFolder.FindChild(destName);
            if (destItem != null)
            {
                if (DelWhenReplace && destItem is IFolder)
                {
                    Folder destItemF = destItem as Folder;
                    if (!destItemF.ClientHasTokenForTree())
                        return new LockedResponse();
                    destItemF.RemoveTree();
                    MoveThisItem(destFolder, destName, parent);
                    return new NoContentResponse();
                }
                else if (destItem is IResource)
                {
                    WebDAVResponse delResp = destItem.Delete();
                    if (!delResp.IsSuccess)
                        return delResp;
                    MoveThisItem(destFolder, destName, parent);
                    return new NoContentResponse();
                }
                else
                    return new ConflictResponse();
            }
            else
            {
                MoveThisItem(destFolder, destName, parent);
                return new CreatedResponse();
            }
        }

        #endregion  // IHierarchyItem Members

        #region IVersionableItem Members

        public IHistory VersionHistory
        {
            get
            {
                SqlCommand cmd = fEngine.CreateNewCommand();
                cmd.CommandText = "SELECT VersionId"
                                  + " FROM Version"
                                  + " WHERE ItemId = @ItemId";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
                if (cmd.ExecuteScalar() != null)
                    return new VersionHistory(this.fEngine, this.fItemId, Path + "?history");

                return null;
            }
        }

        public WebDAVResponse CheckIn(out string newVersionUrl)
        {
            if (!ClientHasToken)
            {
                newVersionUrl = String.Empty;
                return new LockedResponse();
            }

            if (this.VersionHistory != null)
            {
                // Create new version. Copy content and properties from this item to new version.
                SqlCommand cmd = fEngine.CreateNewCommand();

                Version version = this.VersionHistory.CurrentVersion as Version;
                Debug.Assert(version != null);
                int newVersionNumber = version.VersionNumber + 1;
                string newVersionPath = Version.CreateVersionPath(this.fPath, newVersionNumber);
                string creator = (String.IsNullOrEmpty(CurrentUserName))
                                     ? "CreatorDisplayName"
                                     : "@CreatorDisplayName";

                // Create new version.
                cmd.CommandText =
                    "SET @Identity = NEWID()"
                    + " INSERT INTO Version"
                    + " (ItemId, VersionId, VersionNumber, Name, Comment,"
                    + " CreatorDisplayName, ContentType, Created)"
                    + " SELECT @ItemId, @Identity, @VersionNumber, Name, Comment, "
                    + creator + ", ContentType, GETUTCDATE()"
                    + " FROM Item"
                    + " WHERE ItemId = @ItemId";

                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
                cmd.Parameters.Add("@VersionNumber", SqlDbType.Int).Value = newVersionNumber;
                if (!(String.IsNullOrEmpty(CurrentUserName)))
                    cmd.Parameters.Add("@CreatorDisplayName", SqlDbType.NVarChar).Value = CurrentUserName;
                SqlParameter idParm = cmd.Parameters.Add("@Identity", SqlDbType.UniqueIdentifier);
                idParm.Direction = ParameterDirection.Output;

                cmd.ExecuteNonQuery();

                cmd.Parameters.Clear();
                cmd.CommandText = "UPDATE Item SET Comment = '' WHERE ItemId = @ItemId";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;

                cmd.ExecuteNonQuery();

                Guid newID = (Guid) idParm.Value;

                // Copy properties to new version
                cmd.CommandText =
                    "INSERT INTO VersionProperty"
                    + " (VersionId, Name, Namespace, PropVal)"
                    + " SELECT @VersionId, Name, Namespace, PropVal"
                    + " FROM Property"
                    + " WHERE ItemID = @ItemID";

                cmd.Parameters.Add("@VersionId", SqlDbType.UniqueIdentifier).Value = newID;
                cmd.ExecuteNonQuery();

                // Copy content to new version
                File.Copy(this.fFilePath, this.fItemDir + "\\" + newVersionNumber);
                newVersionUrl = newVersionPath;
            }
            else
                newVersionUrl = this.Path;

            SetResourceCheckedOut(false);
            return new OkResponse();
        }

        public WebDAVResponse CheckOut()
        {
            if (!ClientHasToken)
                return new LockedResponse();
            
            SetResourceCheckedOut(true);
            return new OkResponse();
        }

        public WebDAVResponse UnCheckOut()
        {
            if (!ClientHasToken)
                return new LockedResponse();

            if (this.VersionHistory != null)
            {
                // Discard changes.
                // Copy content and properties from current version to this item. Mark item as checked in.
                Version version = (Version) this.VersionHistory.CurrentVersion;

                // Restore properties.
                SqlCommand cmd = fEngine.CreateNewCommand();

                cmd.CommandText = "DELETE FROM Property WHERE ItemID = @ItemID";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
                cmd.ExecuteNonQuery();

                cmd.CommandText = "INSERT INTO Property"
                                  + " (ItemId, Name, Namespace, PropVal)"
                                  + " SELECT @ItemId, Name, Namespace, PropVal"
                                  + " FROM VersionProperty"
                                  + " WHERE VersionId = @VersionId";

                cmd.Parameters.Add("@VersionId", SqlDbType.UniqueIdentifier).Value = version.VersionId;
                cmd.ExecuteNonQuery();

                // Restore content.
                File.Copy(version.FilePath, this.fFilePath, true);
            }
            // Mark item as checked in.
            SetResourceCheckedOut(false);
            return new OkResponse();
        }

        public WebDAVResponse UpdateToVersion(IVersion version)
        {
            if (!ClientHasToken)
                return new LockedResponse();
            // Create a new version and copy content and properties from IVersion passed as a parameter. 
            // Replace content and properties of this item.
            Version ver = version as Version;
            if (ver.ItemId != this.fItemId)
                return new WebDAVResponse(409, "Conflict");

            SqlCommand cmd = fEngine.CreateNewCommand();

            Version currVersion = this.VersionHistory.CurrentVersion as Version;
            int newVersionNumber = currVersion.VersionNumber + 1;

            // Create a new version
            cmd.CommandText =
                "SET @Identity = NEWID()"
                + " INSERT INTO Version"
                + " (ItemId, VersionId, VersionNumber, Name, Comment, CreatorDisplayName, ContentType, Created)"
                +
                " SELECT @ItemId, @Identity, @VersionNumber, Name, Comment, CreatorDisplayName, ContentType, GETUTCDATE()"
                + " FROM Item"
                + " WHERE ItemId = @ItemId";

            cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
            cmd.Parameters.Add("@VersionNumber", SqlDbType.Int).Value = newVersionNumber;
            SqlParameter idParm = cmd.Parameters.Add("@Identity", SqlDbType.UniqueIdentifier);
            idParm.Direction = ParameterDirection.Output;

            cmd.ExecuteNonQuery();
            Guid newID = (Guid) idParm.Value;

            cmd.Parameters.Clear();

            // Copy properties to this item
            cmd.CommandText = "DELETE FROM Property WHERE ItemID = @ItemID";
            cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
            cmd.ExecuteNonQuery();

            cmd.CommandText = "INSERT INTO Property"
                              + " (ItemId, Name, Namespace, PropVal)"
                              + " SELECT @ItemId, Name, Namespace, PropVal"
                              + " FROM VersionProperty"
                              + " WHERE VersionId = @VersionId";

            cmd.Parameters.Add("@VersionId", SqlDbType.UniqueIdentifier).Value = ver.VersionId;
            cmd.ExecuteNonQuery();

            // Copy properties to new version
            cmd.CommandText = "INSERT INTO VersionProperty"
                              + " (VersionId, Name, Namespace, PropVal)"
                              + " SELECT @NewVerId, Name, Namespace, PropVal"
                              + " FROM VersionProperty"
                              + " WHERE VersionId = @VersionId";

            cmd.Parameters.Clear();
            cmd.Parameters.Add("@VersionId", SqlDbType.UniqueIdentifier).Value = ver.VersionId;
            cmd.Parameters.Add("@NewVerId", SqlDbType.UniqueIdentifier).Value = newID;
            cmd.ExecuteNonQuery();

            // Copy content to this item
            File.Copy(ver.FilePath, this.fFilePath, true);

            // Copy content to new version
            File.Copy(ver.FilePath, this.fItemDir + "\\" + newVersionNumber);

            return new OkResponse();
        }

        public AutoVersion AutoVersion
        {
            get
            {
                AutoVersion autoVersion;

                SqlCommand cmd = fEngine.CreateNewCommand();
                cmd.CommandText = "SELECT AutoVersion FROM Item WHERE ItemId = @ItemId";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = ItemId;
                autoVersion = (AutoVersion) cmd.ExecuteScalar();

                return autoVersion;
            }
            set
            {
                SqlCommand cmd = fEngine.CreateNewCommand();
                cmd.CommandText = "UPDATE Item SET AutoVersion = @AutoVersion"
                                  + " WHERE ItemId = @ItemId";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
                cmd.Parameters.Add("@AutoVersion", SqlDbType.Int).Value = (Int32) value;
                cmd.ExecuteNonQuery();
            }
        }

        public bool CheckedOut
        {
            get { return GetDbField("CheckedOut", false); }
        }

        public WebDAVResponse PutUnderVersionControl(bool enable)
        {
            if (enable && this.VersionHistory == null)
            {
                // Create new version. The content and properties of the new version is being copied from this item.
                this.AutoVersion = fAutoVersionMode;

                SqlCommand cmd = fEngine.CreateNewCommand();

                cmd.CommandText =
                    "SET @Identity = NEWID()"
                    + " INSERT INTO Version"
                    + " (ItemId, VersionId, VersionNumber, Name, Comment, CreatorDisplayName, ContentType, Created)"
                    + " SELECT @ItemId, @Identity, 1, Name, Comment, @CreatorDisplayName, ContentType, GETUTCDATE()"
                    + " FROM Item"
                    + " WHERE ItemId = @ItemId";

                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
                cmd.Parameters.Add("@CreatorDisplayName", SqlDbType.NVarChar).Value = CurrentUserName;
                SqlParameter idParam = cmd.Parameters.Add("@Identity", SqlDbType.UniqueIdentifier);
                idParam.Direction = ParameterDirection.Output;

                cmd.ExecuteNonQuery();

                Guid newID = (Guid) idParam.Value;

                cmd.CommandText =
                    "INSERT INTO VersionProperty"
                    + " (VersionId, Name, Namespace, PropVal)"
                    + " SELECT @VersionId, Name, Namespace, PropVal"
                    + " FROM Property"
                    + " WHERE ItemID = @ItemID";

                cmd.Parameters.Clear();
                cmd.Parameters.Add("@VersionId", SqlDbType.UniqueIdentifier).Value = newID;
                cmd.Parameters.Add("@ItemID", SqlDbType.UniqueIdentifier).Value = this.fItemId;
                cmd.ExecuteNonQuery();

                this.SetResourceCheckedOut(false);

                File.Copy(this.fFilePath, this.fItemDir + "\\1");
            }
            else if (!enable)
            {
                // Delete all versions
                SqlCommand cmd = fEngine.CreateNewCommand();

                cmd.CommandText = "DELETE FROM Version WHERE ItemId = @ItemId";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
                cmd.ExecuteNonQuery();

                string[] files = Directory.GetFiles(this.fItemDir);
                foreach (string fileName in files)
                    if (!fileName.EndsWith("current"))
                        File.Delete(fileName);

                return new NoContentResponse();
            }
            return new OkResponse();
        }

        public bool AutoCheckIn
        {
            set { SetDbField("CheckinDuringUnlock", SqlDbType.Bit, value); }
            get { return GetDbField("CheckinDuringUnlock", false); }            
        }

        #endregion // IVersionableItem Members

        #region IDeltaVItem Members

        public string Comment
        {
            get { return GetDbField<string>("Comment", null); }
            set { SetDbField("Comment", SqlDbType.NVarChar, value); }
        }

        public string CreatorDisplayName
        {
            get { return GetDbField<string>("CreatorDisplayName", null); }
            set { SetDbField("CreatorDisplayName", SqlDbType.NVarChar, value); }
        }

        #endregion // IDeltaVItem Members

        #region IResumableUpload Members

        public WebDAVResponse SaveFromStream(long startIndex, long totalContentLength, Stream segment, string contentType)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            long read = 0;
            try
            {
                read = WriteContent(segment, contentType, totalContentLength, startIndex);
            }
            catch (HttpListenerException)
            {
                //We came here because there were troubles with connection.
                //We want to keep content but return ResumeIncompleteResponse,
                //so engine knows we havn't finished uploading.
            }

            if (startIndex + read < totalContentLength)
            {
                return new ResumeIncompleteResponse();
            }
            else if (read > 0)
            {
                return new OkResponse();
            }
            else
            {
                return new NoContentResponse();
            }
        }

        public WebDAVResponse CancelUpload()
        {
            if (!ClientHasToken)
            {
                return new LockedResponse();
            }
            return new OkResponse();
        }

        public DateTime LastChunkSaved
        {
            get
            {
                if (File.Exists(fFilePath))
                {
                    FileInfo file = new FileInfo(fFilePath);
                    return file.LastWriteTimeUtc;
                }
                else
                {
                    return DateTime.MinValue;
                }
            }
        }

        public long BytesUploaded
        {
            get { return ContentLength; }
        }
        
        public long TotalContentLength
        {
            get { return GetDbField<long>("TotalContentLength", -1); }
        }

        public bool CheckInOnFileComplete
        {
            set { SetDbField("CheckinOnFileComplete", SqlDbType.Bit, value); }
            get { return GetDbField("CheckinOnFileComplete", false); }
        }

        #endregion // IResumableUpload Members

        #region IUploadProgress Members

        IResumableUpload[] IUploadProgress.UploadProgress
        {
            get
            {
                IResumableUpload[] items = new IResumableUpload[1];
                items[0] = this;
                return items;
            }
        }

        #endregion // IUploadProgress Members

        #region Internal Members

        internal long WriteContent(Stream content, string contentType, long totalLength,
                                long startIndex)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();
            long bytes = 0;

            if (!Directory.Exists(fItemDir))
            {
                Directory.CreateDirectory(fItemDir);
            }


            int bufSize = 1048576; // 1Mb

            cmd.CommandText = "UPDATE Item SET"
                              + " Modified = GETUTCDATE(),"
                              + " TotalContentLength = @TotalContentLength, "
                              + " ContentType = @ContentType"
                              + " WHERE ItemId = @ItemId;";

            cmd.Parameters.AddWithValue("@ContentType", (object)contentType ?? DBNull.Value);
            cmd.Parameters.AddWithValue("@ItemId", ItemId);
            cmd.Parameters.AddWithValue("@TotalContentLength", totalLength);
            cmd.ExecuteNonQuery();

            if (startIndex == 0)
            {
                File.Delete(fFilePath);
            }

            FileInfo file = new FileInfo(fFilePath);
            using (FileStream fileStream = file.Open(FileMode.OpenOrCreate, FileAccess.Write))
            {
                if (fileStream.Length < startIndex)
                {
                    return 0;
                }
                fileStream.Seek(startIndex, SeekOrigin.Begin);

                byte[] buffer = new byte[bufSize];

                int lastBytesRead;
                while ((lastBytesRead = content.Read(buffer, 0, bufSize)) > 0)
                {
                    fileStream.Write(buffer, 0, lastBytesRead);
                    fileStream.Flush();
                    bytes += lastBytesRead;
                }
            }

            return bytes;
        }

        /// <summary>
        /// Is called if content is empty.
        /// </summary>
        internal void WriteEmptyContent(string contentType)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();

            if (!Directory.Exists(fItemDir))
            {
                Directory.CreateDirectory(fItemDir);
            }

            cmd.CommandText = "UPDATE Item SET"
                                  + " Modified = GETUTCDATE(),"
                                  + " ContentType = @ContentType,"
                                  + " TotalContentLength = 0"
                                  + " WHERE ItemId = @ItemId";

            cmd.Parameters.Add("@ContentType", SqlDbType.NVarChar).Value = (object)contentType ?? DBNull.Value;
            cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = ItemId;
            cmd.ExecuteNonQuery();

            if (File.Exists(fFilePath))
            {
                File.Delete(fFilePath);
            }

            FileStream fileStream = new FileStream(fFilePath, FileMode.CreateNew);
            fileStream.Close();
        }

        internal void RemoveFromDisk()
        {
            if (Directory.Exists(fItemDir))
            {
                string[] files = Directory.GetFiles(fItemDir);

                foreach (string file in files)
                    File.Delete(file);

                Directory.Delete(fItemDir);
            }
        }

        #endregion // Internal Members

        #region Private Members

        private void SetResourceCheckedOut(bool value)
        {
            SetDbField("CheckedOut", SqlDbType.Bit, value);
        }

        private T GetDbField<T>(string columnName, T defaultValue)
        {
            T result = defaultValue;

            SqlCommand cmd = fEngine.CreateNewCommand();
            cmd.CommandText = String.Format("SELECT {0} FROM Item WHERE ItemId = @ItemId", columnName);
            cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
            object obj = cmd.ExecuteScalar();
            if (obj != null && !(obj is DBNull))
                result = (T)obj;

            return result;
        }

        private void SetDbField(string columnName, SqlDbType type, object value)
        {
            SqlCommand cmd = fEngine.CreateNewCommand();
            cmd.CommandText = String.Format("UPDATE Item SET {0} = @value WHERE ItemId = @ItemId", columnName);
            cmd.Parameters.Add("@value", type).Value = value;
            cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = this.fItemId;
            cmd.ExecuteNonQuery();
        }

        #endregion // Private Members

    }
}
