using System;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Security.Principal;
using ITHit.WebDAV.Server;

namespace WebDAVServer.DeltaV
{
    public class WDEngine : Engine
    {
        private static readonly string fConnStr = ConfigurationManager.AppSettings["SqlConnectionString"];
        private readonly Guid fRootParentId = Guid.Empty;
        private SqlConnection fConnection;
        private SqlTransaction fTransaction;
        private readonly IPrincipal user;

        #region Engine Implementation

        public WDEngine(IPrincipal user)
        {
            this.user = user;
        }

        public override IHierarchyItem GetHierarchyItem(string path)
        {
            Guid newItemId = Guid.Empty;
            ItemType itemType = ItemType.Unknown;
            string itemName = string.Empty;
            DateTime itemCreated = DateTime.MinValue;
            DateTime itemModified = DateTime.MinValue;
            Guid versionId = Guid.Empty;
            int versionNumber = 0;

            SqlCommand cmd = CreateNewCommand();

            if (path.Contains("?version=")) // IVersion
            {
                string versionNum = path.Substring(path.LastIndexOf('=') + 1);
                string itemPath = path.Remove(path.IndexOf('?'));

                Resource item = (Resource) GetHierarchyItem(itemPath);

                cmd.CommandText = "SELECT ItemId, VersionId, VersionNumber, Name, Created"
                                  + " FROM Version"
                                  + " WHERE ItemId = @ItemId"
                                  + " AND VersionNumber = '" + versionNum + "'";
                cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = item.ItemId;

                using (SqlDataReader reader = cmd.ExecuteReader())
                {
                    if (reader != null && reader.Read())
                    {
                        newItemId = item.ItemId;
                        itemType = ItemType.Version;
                        versionId = reader.GetGuid(reader.GetOrdinal("VersionId"));
                        versionNumber = reader.GetInt32(reader.GetOrdinal("VersionNumber"));
                        itemName = reader.GetString(reader.GetOrdinal("Name"));
                        itemCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                    }
                }
            }
            else if (path.EndsWith("?history")) // IHistory
            {
                string itemPath = path.Remove(path.IndexOf('?'));
                IHierarchyItem item = GetHierarchyItem(itemPath);

                if (item != null)
                {
                    newItemId = (item as Resource).ItemId;

                    cmd.CommandText = "SELECT VersionId"
                                      + " FROM Version"
                                      + " WHERE ItemId = @ItemId";
                    cmd.Parameters.Add("@ItemId", SqlDbType.UniqueIdentifier).Value = newItemId;

                    using (SqlDataReader reader = cmd.ExecuteReader())
                    {
                        if (reader != null && reader.Read())
                            itemType = ItemType.VersionHistory;
                    }
                }
            }
            else
            {
                int ind = path.IndexOf('?');
                if (ind > 0)
                    path = path.Remove(ind);

                if (path == "/")
                {
                    // get root folder
                    cmd.CommandText = "SELECT ItemId, ItemType, Name, Created, Modified"
                                      + " FROM Item"
                                      + " WHERE ParentItemId = @rootParentId";
                    cmd.Parameters.Add("@rootParentId", SqlDbType.UniqueIdentifier).Value = fRootParentId;

                    using (SqlDataReader reader = cmd.ExecuteReader())
                    {
                        if (reader != null && reader.Read())
                        {
                            newItemId = reader.GetGuid(reader.GetOrdinal("ItemId"));
                            itemType = (ItemType) reader.GetInt32(reader.GetOrdinal("ItemType"));
                            itemName = reader.GetString(reader.GetOrdinal("Name"));
                            itemCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                            itemModified = reader.GetDateTime(reader.GetOrdinal("Modified"));
                        }
                    }
                }
                else
                {
                    // find root
                    cmd.CommandText = "SELECT ItemId FROM Item WHERE ParentItemId = @rootParentId";
                    cmd.Parameters.Add("@rootParentId", SqlDbType.UniqueIdentifier).Value = fRootParentId;
                    Guid id = (Guid) cmd.ExecuteScalar();

                    string[] names = path.Split('/');
                    int last = names.Length - 1;
                    while (last > 0 && names[last] == "") last--;
                    // search for item by path
                    cmd.CommandText =
                        "SELECT ItemId FROM Item"
                        + " WHERE Name = @Name"
                        + " AND ParentItemId = @Parent";
                    cmd.Parameters.Add("@Name", SqlDbType.NVarChar, 255);
                    cmd.Parameters.Add("@Parent", SqlDbType.UniqueIdentifier, 16);
                    cmd.Prepare();
                    for (int i = 0; i < last; i++)
                        if (names[i] != "")
                        {
                            cmd.Parameters["@Name"].Value = names[i];
                            cmd.Parameters["@Parent"].Value = id;
                            object result = cmd.ExecuteScalar();

                            if (result != null)
                            {
                                id = (Guid) result;
                            }
                        }
                    // get item properties
                    cmd.CommandText = "SELECT ItemId, ItemType, Name, Created, Modified"
                                      + " FROM Item"
                                      + " WHERE Name = @Name"
                                      + " AND ParentItemId = @Parent";
                    cmd.Parameters["@Name"].Value = names[last];
                    cmd.Parameters["@Parent"].Value = id;

                    using (SqlDataReader reader = cmd.ExecuteReader())
                        if (reader != null && reader.Read())
                        {
                            newItemId = reader.GetGuid(reader.GetOrdinal("ItemId"));
                            itemType = (ItemType) reader.GetInt32(reader.GetOrdinal("ItemType"));
                            itemName = reader.GetString(reader.GetOrdinal("Name"));
                            itemCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                            itemModified = reader.GetDateTime(reader.GetOrdinal("Modified"));
                        }
                }
            }

            if (newItemId == fRootParentId)
                return null;

            switch (itemType)
            {
                case ItemType.Resource:
                    return new Resource(this, newItemId, itemName, path, itemCreated, itemModified);
                case ItemType.Folder:
                    if (!path.EndsWith("/")) path += "/";
                    return new Folder(this, newItemId, itemName, path, itemCreated, itemModified);
                case ItemType.LockNull:
                    return new LockNull(this, newItemId, itemName, path, itemCreated, itemModified);
                case ItemType.Version:
                    return new Version(this, versionId, newItemId, versionNumber, itemName, path,
                                       itemCreated);
                case ItemType.VersionHistory:
                    return new VersionHistory(this, newItemId, path);
                default:
                    return null;
            }
        }

        #endregion // Engine Implementation

        #region Public Methods

        public SqlCommand CreateNewCommand()
        {
            if (this.fConnection == null)
            {
                this.fConnection = new SqlConnection(fConnStr);
                this.fConnection.Open();
                this.fTransaction = this.fConnection.BeginTransaction(IsolationLevel.ReadUncommitted);
            }
            SqlCommand newCmd = fConnection.CreateCommand();
            newCmd.Transaction = fTransaction;
            return newCmd;
        }

        public void CommitTransaction()
        {
            if (fTransaction != null)
                fTransaction.Commit();
        }

        public void RollBackTransaction()
        {
            if (fTransaction != null)
                fTransaction.Rollback();
        }

        public void CloseConnection()
        {
            if (fConnection != null && fConnection.State != ConnectionState.Closed)
                fConnection.Close();
        }

        public IPrincipal User
        {
            get { return user; }
        }

        #endregion // Public Methods
    }
}