using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.IO;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;

namespace WebDAVServer.NtfsStorage
{
    public class Folder : HierarchyItem, IFolder, IFolderLock
    {
        public Folder(DirectoryInfo directory, WDEngine engine)
            : base(directory, engine)
        {
        }

        #region IFolder Members

        public IHierarchyItem[] Children
        {
            get
            {
                List<IHierarchyItem> children = new List<IHierarchyItem>();

                DirectoryInfo directory = (DirectoryInfo) fileSystemInfo;
                FileSystemInfo[] fileInfos = directory.GetFileSystemInfos();
                foreach (FileSystemInfo item in fileInfos)
                {
                    if (item is DirectoryInfo)
                        children.Add(new Folder((DirectoryInfo)item, engine));
                    else
                        children.Add(new Resource((FileInfo)item, engine));
                }

                return children.ToArray();
            }
        }

        public WebDAVResponse CreateResource(string name)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            string fileName = System.IO.Path.Combine(fileSystemInfo.FullName, name);
            using (new FileStream(fileName, FileMode.CreateNew))
            {
            }

            return new CreatedResponse();
        }

        public WebDAVResponse CreateFolder(string name)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            DirectoryInfo directory = (DirectoryInfo) fileSystemInfo;
            directory.CreateSubdirectory(name);
            return new CreatedResponse();
        }

        #endregion // IFolder Members

        #region IFolderLock Members

        public WebDAVResponse CreateLockNull(string name, ref LockInfo lockInfo)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            string fileName = System.IO.Path.Combine(fileSystemInfo.FullName, name);
            File.Create(fileName).Close();
            Resource newItem = new Resource(new FileInfo(fileName), engine);
            newItem.Lock(ref lockInfo);
            return new CreatedResponse();
        }

        #endregion

        #region HierarchyItem Members Override

        public override WebDAVResponse Delete()
        {            
            StringCollection clientLockTokens = engine.Request.ClientLockTokens;

            ItemResponse[] responses = DeletePartial(this, clientLockTokens).ToArray();

            if (responses.Length > 0)
            {
                MultistatusResponse response = new MultistatusResponse();
                response.AddResponses(responses);
                return response;
            }
            else
                return new NoContentResponse();
        }

        public override WebDAVResponse MoveTo(IFolder folder, string destName)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            Folder targetFolder = (Folder) folder;

            string newDirPath = System.IO.Path.Combine(targetFolder.GetFullPath(),
                                                       destName);
            string newHierDir = targetFolder.GetHierarchyPath() + '/' + destName;
            MultistatusResponse multistatusResponse = new MultistatusResponse();

            WebDAVResponse moveResponse;

            if(File.Exists(newDirPath))
            {
                File.Delete(newDirPath);
                moveResponse = folder.CreateFolder(destName);
                if (!moveResponse.IsSuccess)
                    multistatusResponse.AddResponses(new ItemResponse(newHierDir, moveResponse));
            }
            if (!Directory.Exists(newDirPath))
            {
                moveResponse = folder.CreateFolder(destName);
                if (!moveResponse.IsSuccess)
                    multistatusResponse.AddResponses(new ItemResponse(newHierDir, moveResponse));
            }
            else
                moveResponse = new NoContentResponse();

            if (moveResponse.IsSuccess)
            {
                DirectoryInfo currentDir = fileSystemInfo as DirectoryInfo;
                multistatusResponse.AddResponses(MoveFoldersStructure(currentDir,
                                                                      new DirectoryInfo(newDirPath)));
            }

            if (multistatusResponse.Responses.Length > 0)
                return multistatusResponse;
            else
                return moveResponse;
        }

        public override WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
        {
            Folder targetFolder = folder as Folder;
            if (targetFolder == null)
                throw new ArgumentNullException();

            string newDirPath = System.IO.Path.Combine(targetFolder.GetFullPath(),
                                                       destName);
            string newHierDir = targetFolder.GetHierarchyPath() + '/' + destName;
            MultistatusResponse response = new MultistatusResponse();

            if (!Directory.Exists(newDirPath))
                response.AddResponses(new ItemResponse(newHierDir,
                                                       folder.CreateFolder(destName)));
            DirectoryInfo currentDir = fileSystemInfo as DirectoryInfo;
            if (currentDir != null)
                if (deep)
                    response.AddResponses(CopyFoldersStructure(currentDir,
                                                               new DirectoryInfo(newDirPath)));
                else
                    response.AddResponses(CopyAllFilesFromFolder(currentDir,
                                                                 new DirectoryInfo(newDirPath)));
            return response;
        }

        #endregion // HierarchyItem Members Override

        #region Copy/Move/Delete Private Methods

        /// <summary>
        /// Copies folders structure (and all the files) from source to destination
        /// (both folders should exist)
        /// </summary>
        /// <param name="sourceFolder">The folder structure will be copied from here</param>
        /// <param name="destFolder">The folder stucture will be copied here</param>
        /// <returns>Array of responses</returns>
        private ItemResponse[] CopyFoldersStructure(DirectoryInfo sourceFolder, DirectoryInfo destFolder)
        {
            List<ItemResponse> listResponse = new List<ItemResponse>();
            DirectoryInfo[] dirs = sourceFolder.GetDirectories();
            listResponse.AddRange(CopyAllFilesFromFolder(sourceFolder, destFolder));
            Folder newFolder = new Folder(destFolder, engine);

            foreach (DirectoryInfo dir in dirs)
            {
                string newDir = System.IO.Path.Combine(destFolder.FullName, dir.Name);
                string newHierDir = newFolder.GetHierarchyPath() + '/' + dir.Name;
                listResponse.Add(new ItemResponse(newHierDir, newFolder.CreateFolder(dir.Name)));
                listResponse.AddRange(CopyFoldersStructure(dir, new DirectoryInfo(newDir)));
            }
            return listResponse.ToArray();
        }

        private ItemResponse[] CopyAllFilesFromFolder(DirectoryInfo sourceFolder, DirectoryInfo destFolder)
        {
            List<ItemResponse> listResponse = new List<ItemResponse>();
            FileInfo[] files = sourceFolder.GetFiles();
            Folder newFolder = new Folder(destFolder, engine);

            foreach (FileInfo file in files)
            {
                Resource res = new Resource(file, engine);
                listResponse.Add(new ItemResponse(res.GetHierarchyPath(),
                    res.CopyTo(newFolder, file.Name, false)));
            }

            return listResponse.ToArray();
        }

        /// <summary>
        /// Moves folders structure (and all the files) from source to destination
        /// (both folders should exist)
        /// </summary>
        /// <param name="sourceFolder">The folder structure will be copied from here</param>
        /// <param name="destFolder">The folder stucture will be copied here</param>
        /// <returns>Array of responses</returns>
        private ItemResponse[] MoveFoldersStructure(DirectoryInfo sourceFolder, DirectoryInfo destFolder)
        {
            List<ItemResponse> listResponse = new List<ItemResponse>();
            DirectoryInfo[] dirs = sourceFolder.GetDirectories();
            listResponse.AddRange(MoveAllFilesFromFolder(sourceFolder, destFolder));
            Folder newFolder = new Folder(destFolder, engine);

            foreach (DirectoryInfo dir in dirs)
            {
                string newDir = System.IO.Path.Combine(destFolder.FullName, dir.Name);
                string newHierDir = newFolder.GetHierarchyPath() + '/' + dir.Name;
                WebDAVResponse resp = newFolder.CreateFolder(dir.Name);
                if (!resp.IsSuccess)
                    listResponse.Add(new ItemResponse(newHierDir, resp));
                listResponse.AddRange(MoveFoldersStructure(dir, new DirectoryInfo(newDir)));
            }

            bool failedToMove = false;
            foreach (ItemResponse itemResponse in listResponse)
                if (!itemResponse.Response.IsSuccess)
                    failedToMove = true;

            if (!failedToMove)
            {
                Folder current = new Folder(sourceFolder, engine);
                WebDAVResponse resp = current.Delete();
                if (!resp.IsSuccess)
                    listResponse.Add(new ItemResponse(current.GetHierarchyPath(), resp));
            }
            return listResponse.ToArray();
        }

        private ItemResponse[] MoveAllFilesFromFolder(DirectoryInfo sourceFolder, DirectoryInfo destFolder)
        {
            List<ItemResponse> listResponse = new List<ItemResponse>();
            FileInfo[] files = sourceFolder.GetFiles();
            Folder newFolder = new Folder(destFolder, engine);

            foreach (FileInfo file in files)
            {
                Resource res = new Resource(file, engine);
                string path = res.GetHierarchyPath();
                WebDAVResponse resp = res.MoveTo(newFolder, file.Name);
                if (!resp.IsSuccess)
                    listResponse.Add(new ItemResponse(path, resp));
            }

            return listResponse.ToArray();
        }

        private List<ItemResponse> DeletePartial(Folder folder, StringCollection clientLockTokens)
        {
            List<ItemResponse> listResponse = new List<ItemResponse>();
            LockInfo[] locks;

            foreach (IHierarchyItem child in folder.Children)
            {
                Folder childFolder = child as Folder;
                if (childFolder != null)
                {
                    listResponse.AddRange(DeletePartial(childFolder, clientLockTokens));
                }
                else
                {
                    HierarchyItem item = (HierarchyItem) child;
                    locks = item.ActiveLocks;
                    if (locks.Length == 0 || ClientHasLocks(locks, clientLockTokens))
                        item.Delete();
                    else
                        listResponse.Add(new ItemResponse(item.GetHierarchyPath(),
                                                          new LockedResponse()));
                }
            }

            locks = folder.ActiveLocks;
            if (locks.Length == 0 || ClientHasLocks(locks, clientLockTokens))
            {
                if (folder.Children.Length == 0)
                    folder.fileSystemInfo.Delete();
            }
            else
                listResponse.Add(new ItemResponse(folder.GetHierarchyPath(),
                                                  new LockedResponse()));
            return listResponse;
        }

        private static bool ClientHasLocks(LockInfo[] locks, StringCollection clientLockTokens)
        {
            foreach (LockInfo lockInfo in locks)
            {
                bool found = false;
                foreach (string lockToken in clientLockTokens)
                    if (lockToken == lockInfo.Token)
                    {
                        found = true;
                        break;
                    }
                if (!found)
                    return false;
            }
            return true;
        }

        #endregion // Copy/Move/Delete Private Methods

    }
}
