using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Configuration;
using System.IO;
using System.Text;
using System.Xml.Serialization;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;
using System.ComponentModel;
using WebDAVServer.NtfsStorage.NtfsStreamsEngine;

namespace WebDAVServer.NtfsStorage
{
    public abstract class HierarchyItem : IHierarchyItem, ILock
    {
        protected FileSystemInfo fileSystemInfo;
        protected static int bufSize = 1048576; //(1Mb) buffer size used when reading and writing file content
        protected WDEngine engine;
        private string repositoryPath = ConfigurationManager.AppSettings["RepositoryPath"].TrimEnd('\\');

        public HierarchyItem(FileSystemInfo fileSystemInfo, WDEngine engine)
        {
            this.fileSystemInfo = fileSystemInfo;
            this.engine = engine;
        }

        #region IHierarchyItem Members

        public string Name
        {
            get { return fileSystemInfo.Name; }
        }

        public DateTime Created
        {
            get { return fileSystemInfo.CreationTimeUtc; }
        }

        public DateTime Modified
        {
            get { return fileSystemInfo.LastWriteTimeUtc; }
        }

        public IFolder Parent
        {
            get
            {
                string parentPath = fileSystemInfo.FullName.TrimEnd('\\');
                int index = parentPath.LastIndexOf('\\');
                if (index <= repositoryPath.Length - 1)
                    return null;
                DirectoryInfo directory = new DirectoryInfo(parentPath.Remove(index));
                return new Folder(directory, engine);
            }
        }

        public virtual WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
        {
            return new NotAllowedResponse();
        }

        public virtual WebDAVResponse MoveTo(IFolder folder, string destName)
        {
            return new NotAllowedResponse();
        }

        public abstract WebDAVResponse Delete();

        public WebDAVResponse GetProperties(ref Property[] props)
        {
            string propXml = GetAlternateStreamText("Properties");

            if (string.IsNullOrEmpty(propXml))
                return new OkResponse();

            List<Property> list = (List<Property>) Deserialize(propXml, typeof (List<Property>));

            if (props == null) // get all properties
            {
                props = list.ToArray();
            }
            else // get selected properties
            {
                foreach (Property p in list)
                    for (int i = 0; i < props.Length; i++)
                        if (p.Name == props[i].Name && p.Namespace == props[i].Namespace)
                        {
                            props[i] = p;
                            break;
                        }
            }
            return new OkResponse();
        }

        public WebDAVResponse GetPropertyNames(ref Property[] props)
        {
            string propXml = GetAlternateStreamText("Properties");

            if (string.IsNullOrEmpty(propXml))
                return new OkResponse();

            List<Property> list = (List<Property>) Deserialize(propXml, typeof (List<Property>));
            props = list.ToArray();
            return new OkResponse();
        }

        public virtual WebDAVResponse UpdateProperties(Property[] setProps, Property[] delProps)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            List<Property> list;
            string propXml = GetAlternateStreamText("Properties");
            list = !string.IsNullOrEmpty(propXml)
                       ? (List<Property>) Deserialize(propXml, typeof (List<Property>))
                       : new List<Property>();

            MultipropResponse resp = new MultipropResponse();

            if (setProps != null)
                foreach (Property p in setProps)
                {
                    bool exist = false;
                    for (int i = 0; i < list.Count;i++ )
                        if (list[i].Name == p.Name &&
                            list[i].Namespace == p.Namespace)
                        {
                            list[i] = p;
                            exist = true;
                            break;
                        }
                    if (!exist)
                        list.Add(p);
                    resp.AddResponses(new PropResponse(p, new OkResponse()));
                }

            if (delProps != null)
                foreach (Property p in delProps)
                {
                    bool found = false;
                    for (int i = 0; i < list.Count; i++)
                        if (list[i].Name == p.Name &&
                            list[i].Namespace == p.Namespace)
                        {
                            list.RemoveAt(i);
                            resp.AddResponses(new PropResponse(p, new OkResponse()));
                            found = true;
                            break;
                        }

                    if (!found)
                        resp.AddResponses(new PropResponse(p, new OkResponse()));
                }

            if ((setProps != null && setProps.Length > 0) ||
                (delProps != null && delProps.Length > 0))
            {
                RewriteStream("Properties", Serialize(list));
            }
            return resp;
        }

        public virtual string Path
        {
            get
            {
                return this.GetHierarchyPath();
            }
        }

        #endregion // IHierarchyItem Members

        #region ILock Members

        /// <exclude/>
        public class Pair
        {
            public Pair()
            { }
            public Pair(LockInfo li, DateTime dt)
            {
                this.LockInfo = li;
                this.DateTime = dt;
            }

            public LockInfo LockInfo;
            public DateTime DateTime;
        }

        public LockInfo[] ActiveLocks
        {
            get
            {
                string propXml = GetAlternateStreamText("Locks");

                if (string.IsNullOrEmpty(propXml))
                    return new LockInfo[0];

                List<Pair> list = (List<Pair>)Deserialize(propXml, typeof(List<Pair>));

                RemoveExpired(ref list);
                LockInfo[] locks = new LockInfo[list.Count];
                
                for (int i = 0; i < locks.Length; i++)
                {
                    LockInfo li = list[i].LockInfo;
                    if (list[i].DateTime == DateTime.MaxValue)
                        li.Timeout = -1;
                    else
                        li.Timeout = (list[i].DateTime.Ticks - DateTime.Now.Ticks)/10000000;
                    locks[i] = li;
                }
                
                return locks;
            }
        }

        public WebDAVResponse Lock(ref LockInfo lockInfo)
        {
            if (ItemHasLock(lockInfo.Shared))
                return new LockedResponse();

            if (lockInfo.Deep && this is IFolder)
                return new NotImplementedResponse();

            lockInfo.Token = Guid.NewGuid().ToString();

            string propXml = GetAlternateStreamText("Locks");

            List<Pair> list = !string.IsNullOrEmpty(propXml)
                                      ? (List<Pair>)Deserialize(propXml, typeof(List<Pair>))
                                      : new List<Pair>();
            RemoveExpired(ref list);
            list.Add(new Pair(lockInfo, lockInfo.Timeout >= 0
                                            ? DateTime.Now.AddSeconds(lockInfo.Timeout)
                                            : DateTime.MaxValue));
            RewriteStream("Locks", Serialize(list));

            return new OkResponse();
        }

        public WebDAVResponse RefreshLock(ref LockInfo lockInfo)
        {
            string propXml = GetAlternateStreamText("Locks");

            List<Pair> locks = !string.IsNullOrEmpty(propXml)
                                   ? (List<Pair>) Deserialize(propXml, typeof (List<Pair>))
                                   : new List<Pair>();
            LockInfo li = new LockInfo();
            int i;

            for (i = 0; i < locks.Count;i++)
                if (lockInfo.Token == locks[i].LockInfo.Token)
                {
                    li = locks[i].LockInfo;
                    break;
                }

            if (i == locks.Count)
                return new PreconditionFailedResponse();

            lockInfo.Owner = li.Owner;
            lockInfo.Shared = li.Shared;
            lockInfo.Deep = li.Deep;

            locks.RemoveAt(i);
            locks.Add(new Pair(lockInfo, lockInfo.Timeout >= 0
                                   ? DateTime.Now.AddSeconds(lockInfo.Timeout)
                                   : DateTime.MaxValue));
            RewriteStream("Locks", Serialize(locks));

            return new OkResponse();
        }

        public WebDAVResponse Unlock(string lockToken)
        {
            string propXml = GetAlternateStreamText("Locks");

            List<Pair> locks = !string.IsNullOrEmpty(propXml)
                                   ? (List<Pair>) Deserialize(propXml, typeof (List<Pair>))
                                   : new List<Pair>();
            int i;

            for (i = 0; i < locks.Count; i++)
                if (locks[i].LockInfo.Token == lockToken)
                    break;

            if (i == locks.Count)
                return new PreconditionFailedResponse();

            // remove lock from existing item
            locks.RemoveAt(i);
            if (locks.Count > 0)
            {
                RemoveExpired(ref locks);
                RewriteStream("Locks", Serialize(locks));
            }
            else
                (new AlternateDataStreamInfo(this.fileSystemInfo.FullName, "Locks")).Delete();

            return new NoContentResponse();
        }

        private static void RemoveExpired(ref List<Pair> list)
        {
            if (list.Count == 0)
                return;

            for (int i = list.Count - 1; i >= 0; i--)
                if (list[i].DateTime <= DateTime.Now)
                    list.RemoveAt(i);
        }

        #endregion //  ILock Members

        #region Internal Methods

        internal string GetFullPath()
        {
            return fileSystemInfo.FullName.TrimEnd('\\');
        }

        /// <summary>
        /// Returns path of the item in the hierarchy tree
        /// </summary>
        internal string GetHierarchyPath()
        {
            int index = repositoryPath.Length;
            if (this.GetFullPath().Length > index) index++;

            return this.GetFullPath().Remove(0, index).Replace('\\', '/');
        }

        internal static string Serialize(object list)
        {
            if (list == null)
                throw new ArgumentNullException("list");

            XmlSerializer sr = new XmlSerializer(list.GetType());
            StringBuilder sb = new StringBuilder();
            using (StringWriter w = new StringWriter(sb, System.Globalization.CultureInfo.InvariantCulture))
            {
                sr.Serialize(w, list);
                return sb.ToString();
            }
        }

        internal static object Deserialize(string xmlString, Type targetType)
        {
            if (string.IsNullOrEmpty(xmlString))
                return null;

            using (StringReader reader = new StringReader(xmlString))
            {
                XmlSerializer dsr = new XmlSerializer(targetType);
                return dsr.Deserialize(reader);
            }
        }

        protected string GetAlternateStreamText(string streamName)
        {
            string propXml = null;
            try
            {
                AlternateDataStreamInfo adsi = new AlternateDataStreamInfo(this.fileSystemInfo.FullName,
                                                                           streamName);             
                using (StreamReader sr = adsi.OpenText())
                {
                    propXml = sr.ReadToEnd();
                }
            }
            catch(Win32Exception)
            {
                // no stream found with name specified
            }
            return propXml;
        }

        protected static void RewriteStream(string filePath, string streamName, string newContent)
        {
            AlternateDataStreamInfo asi = new AlternateDataStreamInfo(filePath, streamName);
            asi.Delete();
            using (FileStream fs = asi.OpenWrite())
            {
                byte[] bytes = Encoding.UTF8.GetBytes(newContent);
                fs.Write(bytes, 0, bytes.Length);
            }
        }

        protected void RewriteStream(string streamName, string newContent)
        {
            RewriteStream(this.fileSystemInfo.FullName, streamName, newContent);
        }

        #endregion // Internal Methods

        #region Lock Methods

        internal bool ClientHasToken
        {
            get
            {
                LockInfo[] itemLocks = ActiveLocks;
                if (itemLocks.Length == 0)
                    return true;
                
                StringCollection clientLockTokens = engine.Request.ClientLockTokens;
                for (int i = 0; i < clientLockTokens.Count; i++)
                    for (int j = 0; j < itemLocks.Length; j++)
                        if (clientLockTokens[i] == itemLocks[j].Token)
                            return true;
                return false;
            }
        }

        protected bool ItemHasLock(bool skipShared)
        {
            LockInfo[] locks = ActiveLocks;
            if (locks.Length == 0)
                return false;
            bool allShared = true;
            foreach (LockInfo lockInfo in locks)
                if (!lockInfo.Shared)
                {
                    allShared = false;
                    break;
                }
            return !(skipShared && allShared);
        }

        #endregion // lock methods

    }
}
