﻿using System;
using System.ComponentModel;
using System.IO;
using System.Runtime.InteropServices;
using Microsoft.Win32.SafeHandles;

namespace WebDAVServer.NtfsStorage.NtfsStreamsEngine
{
	/// <summary>
	/// Safe native methods.
	/// </summary>
	internal static class SafeNativeMethods
	{
		#region Constants and flags

		public const int MaxPath = 256;
		private const string longPathPrefix = @"\\?\";
		public const char StreamSeparator = ':';
		public const int DefaultBufferSize = 0x1000;

		private const int errorFileNotFound = 2;

		[Flags]
		public enum NativeFileFlags : uint
		{
			WriteThrough		= 0x80000000,
			Overlapped			= 0x40000000,
			NoBuffering			= 0x20000000,
			RandomAccess		= 0x10000000,
			SequentialScan		= 0x8000000,
			DeleteOnClose		= 0x4000000,
			BackupSemantics		= 0x2000000,
			PosixSemantics		= 0x1000000,
			OpenReparsePoint	= 0x200000,
			OpenNoRecall		= 0x100000
		}

		[Flags]
		public enum NativeFileAccess : uint
		{
			GenericRead		= 0x80000000,
			GenericWrite	= 0x40000000
		}

		#endregion

		#region P/Invoke Methods

		[DllImport("kernel32.dll")]
		private static extern int GetFileType(SafeFileHandle handle);

		[DllImport("kernel32", CharSet = CharSet.Unicode, SetLastError = true)]
		private static extern SafeFileHandle CreateFile(
			string name,
			NativeFileAccess access,
			FileShare share,
			IntPtr security,
			FileMode mode,
			NativeFileFlags flags,
			IntPtr template);

		[DllImport("kernel32", CharSet = CharSet.Unicode, SetLastError = true)]
		[return: MarshalAs(UnmanagedType.Bool)]
		private static extern bool DeleteFile(string name);

        [DllImport("kernel32.dll")]
        private static extern int SetEndOfFile(SafeFileHandle handle);

        [DllImport("kernel32.dll")]
        private static extern bool SetFileValidData(SafeFileHandle handle, long dataLength);

		#endregion

		#region Utility Methods

		public static void ThrowLastWin32Error()
		{
			int error = Marshal.GetLastWin32Error();
			if (0 != error)
			{
				int hr = Marshal.GetHRForLastWin32Error();
				if (0 > hr) Marshal.ThrowExceptionForHR(error);
				throw new Win32Exception(error);
			}
		}

        public static NativeFileAccess ToNative(FileAccess access)
        {
            NativeFileAccess result = 0;
            if (FileAccess.Read == (FileAccess.Read & access)) result |= NativeFileAccess.GenericRead;
            if (FileAccess.Write == (FileAccess.Write & access)) result |= NativeFileAccess.GenericWrite;
            return result;
        }

		public static string BuildStreamPath(string filePath, string streamName)
		{
			string result = filePath.TrimEnd('\\');
			if (!string.IsNullOrEmpty(filePath))
			{
				if (1 == result.Length) result = ".\\" + result;
				result += StreamSeparator + streamName;
				if (MaxPath <= result.Length) result = longPathPrefix + result;
			}
			return result;
		}

		public static bool SafeDeleteFile(string name)
		{
			if (string.IsNullOrEmpty(name)) throw new ArgumentNullException("name");

			bool result = DeleteFile(name);
			if (!result)
			{
				int errorCode = Marshal.GetLastWin32Error();
				if (errorFileNotFound != errorCode) ThrowLastWin32Error();
			}

			return result;
		}

		public static SafeFileHandle SafeCreateFile(string path, NativeFileAccess access, FileShare share, IntPtr security, FileMode mode, NativeFileFlags flags, IntPtr template)
		{
			SafeFileHandle result = CreateFile(path, access, share, security, mode, flags, template);
			if (!result.IsInvalid && 1 != GetFileType(result))
			{
				result.Dispose();
				throw new NotSupportedException(string.Format("The specified file name '{0}' is not a disk-based file.", path));
			}

			return result;
		}
        
        public static void TruncateFile(FileInfo file)
        {
            // There is no another way to truncate file stream
            using (FileStream fs = file.OpenWrite())
            {
                SetFileValidData(fs.SafeFileHandle, 0);
                SetEndOfFile(fs.SafeFileHandle);
            }
        }

		#endregion
	}
}
