using System;
using System.Collections.Generic;
using System.IO;
using System.Web;
using ITHit.WebDAV.Logger;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;
using ITHit.WebDAV.Server.ResumableUpload;
using WebDAVServer.NtfsStorage.NtfsStreamsEngine;

namespace WebDAVServer.NtfsStorage
{
    public class Resource : HierarchyItem, IResource, IResumableUpload, IUploadProgress
    {
        public Resource(FileInfo file, WDEngine engine)
            : base(file, engine)
        {
        }

        #region IResource Members

        public string ContentType
        {
            get
            {
                return MimeType.GetMimeType(this.fileSystemInfo.Extension)
                        ?? "application/octet-stream";
            }
        }

        public long ContentLength
        {
            get
            {
                FileInfo file = (FileInfo) this.fileSystemInfo;
                return file.Length;
            }
        }

        public WebDAVResponse WriteToStream(Stream output, long startIndex, long count)
        {
            FileInfo file = (FileInfo) this.fileSystemInfo;
            byte[] buffer = new byte[bufSize];

            using (FileStream fileStream = file.Open(FileMode.Open, FileAccess.Read, FileShare.Read))
            {
                fileStream.Seek(startIndex, SeekOrigin.Begin);
                int bytesRead;
                while ((bytesRead = fileStream.Read(buffer, 0, (int)(count > bufSize ? bufSize : count))) > 0)
                {
                    try
                    {
                        output.Write(buffer, 0, bytesRead);
                    }
                    catch (HttpException)
                    {
                        // The remote host closed the connection
                        return new OkResponse();
                    }
                    count -= bytesRead;
                }
            }

            return new OkResponse();
        }

        public WebDAVResponse SaveFromStream(Stream content, string contentType)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            FileInfo file = (FileInfo)fileSystemInfo;
            RewriteStream("TotalContentLength", content.Length.ToString());

            byte[] buffer = new byte[bufSize];
            using (FileStream fileStream = file.Open(FileMode.Truncate, FileAccess.Write))
            {
                int bytesRead;
                while ((bytesRead = content.Read(buffer, 0, bufSize)) > 0)
                    fileStream.Write(buffer, 0, bytesRead);
            }

            return new OkResponse();
        }
        
        #endregion // IResource Members

        #region HierarchyItem Members Override

        public override WebDAVResponse Delete()
        {
            if (!ClientHasToken)
                return new LockedResponse();

            fileSystemInfo.Delete();
            return new NoContentResponse();
        }

        public override WebDAVResponse MoveTo(IFolder folder, string destName)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            Folder targetFolder = (Folder) folder;

            if (!Directory.Exists(targetFolder.GetFullPath()))
                return new ConflictResponse();

            string newDirPath = System.IO.Path.Combine(targetFolder.GetFullPath(),
                                                       destName);
            try
            {
                if (!File.Exists(newDirPath))
                {
                    File.Move(fileSystemInfo.FullName, newDirPath);
                    return new CreatedResponse();
                }
                else
                {
                    File.Delete(newDirPath);
                    File.Move(fileSystemInfo.FullName, newDirPath);
                    return new NoContentResponse();
                }
            }
            catch (IOException e)
            {
                FileLogger.WriteMessage(e.ToString());
                return new ConflictResponse();
            }
        }

        public override WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
        {
            Folder targetFolder = (Folder) folder;

            if (targetFolder == null || !Directory.Exists(targetFolder.GetFullPath()))
                return new ConflictResponse();

            string newDirPath = System.IO.Path.Combine(targetFolder.GetFullPath(),
                                                       destName);
            try
            {
                if (File.Exists(newDirPath))
                {
                    Resource res = new Resource(new FileInfo(newDirPath), engine);
                    if (!res.ClientHasToken)
                        return new LockedResponse();
                    File.Delete(newDirPath);
                    File.Copy(fileSystemInfo.FullName, newDirPath);
                    // Removing possible duplicated locks
                    (new AlternateDataStreamInfo(newDirPath, "Locks")).Delete();
                    return new NoContentResponse();
                }
                else if (Directory.Exists(newDirPath))
                {
                    Directory.Delete(newDirPath);
                    File.Copy(fileSystemInfo.FullName, newDirPath);
                    // Removing possible duplicated locks
                    (new AlternateDataStreamInfo(newDirPath, "Locks")).Delete();
                    return new NoContentResponse();
                }
                else
                {
                    File.Copy(fileSystemInfo.FullName, newDirPath);
                    // Removing possible duplicated locks
                    (new AlternateDataStreamInfo(newDirPath, "Locks")).Delete();
                    return new CreatedResponse();
                }
            }
            catch (IOException e)
            {
                FileLogger.WriteMessage(e.ToString());
                return new ConflictResponse();
            }
        }

        #endregion // HierarchyItem Members Override

        #region IResumableUpload Members

        public WebDAVResponse SaveFromStream(long startIndex, long totalContentLength, Stream content, string contentType)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            long read = WriteContent(content, contentType, totalContentLength, startIndex);

            if (startIndex + read < totalContentLength)
                return new ResumeIncompleteResponse(); // more segments expected

            if (read > 0)
                return new OkResponse();
            else
                return new NoContentResponse();
        }

        public WebDAVResponse CancelUpload()
        {
            if (!ClientHasToken)
                return new LockedResponse();

            // Client do not plan to restore upload. Remove any temporary files / cleanup resources here.

            return new OkResponse();
        }

        public DateTime LastChunkSaved
        {
            get
            {
                if (File.Exists(this.fileSystemInfo.FullName))
                {
                    FileInfo file = new FileInfo(this.fileSystemInfo.FullName);
                    return file.LastWriteTimeUtc;
                }
                else
                    return DateTime.MinValue;
            }
        }

        public long BytesUploaded
        {
            get { return ContentLength; }
        }

        public long TotalContentLength
        {
            get
            {
                string value = GetAlternateStreamText("TotalContentLength");
                if (value == null)
                    return 0;
                else
                    return long.Parse(value);
            }
        }

        public bool CheckInOnFileComplete
        {
            set
            {
                RewriteStream("CheckinOnFileComplete", value.ToString());
            }
            get
            {
                string value = GetAlternateStreamText("CheckinOnFileComplete");
                if (value == null)
                    return false;
                else
                    return bool.Parse(value);
            }
        }

        #endregion // IResumableUpload Members

        #region IUploadProgress Members

        public IResumableUpload[] UploadProgress
        {
            get
            {
                List<IResumableUpload> items = new List<IResumableUpload>();
                items.Add(this);
                return items.ToArray();
            }
        }

        #endregion

        public long WriteContent(Stream content, string contentType, long totalLength,
                                long startIndex)
        {
            HttpContext.Current.Server.ScriptTimeout = int.MaxValue;
            long bytes = 0;

            FileInfo file = (FileInfo) this.fileSystemInfo;
            string fileName = file.FullName;

            if (startIndex == 0 && file.Length > 0)
                SafeNativeMethods.TruncateFile(file);

            RewriteStream(fileName, "TotalContentLength", totalLength.ToString());

            using (FileStream fileStream = file.Open(FileMode.OpenOrCreate, FileAccess.Write))
            {
                if (fileStream.Length < startIndex)
                    return 0;

                fileStream.Seek(startIndex, SeekOrigin.Begin);
                byte[] buffer = new byte[bufSize];

                int lastBytesRead;
                while ((lastBytesRead = content.Read(buffer, 0, bufSize)) > 0)
                {
                    fileStream.Write(buffer, 0, lastBytesRead);
                    fileStream.Flush();
                    bytes += lastBytesRead;
                }
            }

            return bytes;
        }
         
    }
}
