using System;
using System.Collections;
using System.Data;
using System.Data.SqlClient;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;

namespace WebDAVServer.SqlStorage
{
    public class Folder : HierarchyItem, IFolder, IFolderLock
    {
        public Folder(int id, string name, string path, DateTime created, DateTime modified, WDEngine engine)
            : base(id, name, path, created, modified, engine)
        {
        }

		#region IFolder Members

		public IHierarchyItem[] Children
		{
			get
			{
				ArrayList children = new ArrayList();

                using (SqlConnection conn = new SqlConnection(connStr))
                {
                    conn.Open();

                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SELECT ID, ItemType, Name, Created, Modified"
                                      + " FROM Repository"
                                      + " WHERE Parent = @Parent";

                    cmd.Parameters.Add("@Parent", SqlDbType.Int).Value = ID;

                    using (SqlDataReader reader = cmd.ExecuteReader())
                        while (reader.Read())
                        {
                            int childID = reader.GetInt32(reader.GetOrdinal("ID"));
                            ItemType childType = (ItemType) reader.GetByte(reader.GetOrdinal("ItemType"));
                            string childName = reader.GetString(reader.GetOrdinal("Name"));
                            DateTime childCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                            DateTime childModified = reader.GetDateTime(reader.GetOrdinal("Modified"));
                            switch (childType)
                            {
                                case ItemType.Resource:
                                    children.Add(
                                        new Resource(childID, childName, path + childName, childCreated, childModified,
                                                     engine));
                                    break;
                                case ItemType.Folder:
                                    children.Add(
                                        new Folder(childID, childName, path + childName + "/", childCreated,
                                                   childModified, engine));
                                    break;
                                case ItemType.LockNull:
                                    children.Add(
                                        new LockNull(childID, childName, path + childName, childCreated, childModified,
                                                     engine));
                                    break;
                            }
                        }
                }

			    return (IHierarchyItem[])children.ToArray(typeof(IHierarchyItem));
			}
		}

		public WebDAVResponse CreateResource(string resourceName)
		{
			if(!ClientHasToken)
				return new LockedResponse();

		    using (SqlConnection conn = new SqlConnection(connStr))
		    {
		        SqlTransaction trans = null;
		        conn.Open();
		        try
		        {
		            trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted);
		            CreateChild(trans, resourceName, ItemType.Resource);
		            trans.Commit();
		        }
		        catch
		        {
		            if(trans != null) trans.Rollback();
		            throw;
		        }
		    }

		    return new CreatedResponse();
		}

		public WebDAVResponse CreateFolder(string folderName)
		{
			if(!ClientHasToken)
				return new LockedResponse();

		    using (SqlConnection conn = new SqlConnection(connStr))
		    {
		        SqlTransaction trans = null;
		        conn.Open();
		        try
		        {
		            trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted);
		            CreateChild(trans, folderName, ItemType.Folder);
		            trans.Commit();
		        }
		        catch
		        {
		            if(trans != null) trans.Rollback();
		            throw;
		        }
		    }
		    return new CreatedResponse();
        }
        
        #endregion // IFolder Members

        #region IFolderLock Members

        public WebDAVResponse CreateLockNull(string newItemName, ref LockInfo lockInfo)
		{
			LockInfo[] locks = ActiveLocks;
			if(locks.Length > 0)
			{
				for(int i=0;i<locks.Length;i++)
					if(locks[i].Deep && (!lockInfo.Shared || !locks[i].Shared))
						return new LockedResponse();

				if(!ClientHasToken)
					return new LockedResponse();
			}
            HierarchyItem newItem;
            using (SqlConnection conn = new SqlConnection(connStr))
            {
                SqlTransaction trans = null;
                conn.Open();
                try
                {
                    trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted);
                    newItem = CreateChild(trans, newItemName, ItemType.LockNull);
                    trans.Commit();
                }
                catch
                {
                    if(trans != null) trans.Rollback();
                    throw;
                }
            }
            newItem.Lock(ref lockInfo);

			return new CreatedResponse();
        }

        #endregion // IFolderLock Members

        #region HierarchyItem Members Override

        public override WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
		{
			Folder destFolder = folder as Folder;
			if(destFolder == null)
				return new ConflictResponse();
            if(!destFolder.ClientHasToken)
                return new LockedResponse();

            IHierarchyItem destItem = destFolder.FindChild(destName);
            Folder newDestFolder;
            
            if(destItem != null)
            {
                if(DelWhenReplace && destItem is IResource)
                {
                    WebDAVResponse delResp = destItem.Delete();
                    if(!delResp.IsSuccess)
                        return delResp;
                    newDestFolder = CopyThisItem(destFolder, null, destName);
                }
                else
                {
                    newDestFolder = destItem as Folder;
                    if(newDestFolder == null)
                        return new ConflictResponse();
                }
            }
            else
            {
                newDestFolder = CopyThisItem(destFolder,null,destName);
            }
            // copy children
            MultistatusResponse mr = new MultistatusResponse();
            if(deep)
            {
                foreach(IHierarchyItem child in Children)
                {
                    HierarchyItem dbchild = (HierarchyItem) child;
                    WebDAVResponse resp = dbchild.CopyTo(newDestFolder, child.Name, deep);
                    if(!resp.IsSuccess)
                        mr.AddResponses(new ItemResponse(dbchild.Path, resp));
                    else
                    {
                        MultistatusResponse mrchild = resp as MultistatusResponse;
                        if(mrchild != null)
                            mr.AddResponses(mrchild.Responses);
                    }
                }
            }

            if(mr.Responses.Length > 0)
                return mr;
            else if(destItem == null)
                return new CreatedResponse();
            else
                return new NoContentResponse();
        }

		public override WebDAVResponse MoveTo(IFolder folder, string destName)
		{
			Folder destFolder = folder as Folder;
			if(destFolder == null)
				return new ConflictResponse();
            
			Folder parent = Parent as Folder;
			if(parent == null)
				return new ConflictResponse();
            
			if(!ClientHasToken || !destFolder.ClientHasToken || !parent.ClientHasToken)
				return new LockedResponse();
			
			HierarchyItem destItem = destFolder.FindChild(destName);
            Folder newDestFolder;

            // copy this folder
            if(destItem != null)
            {
                if(DelWhenReplace && destItem is IResource)
                {
                    WebDAVResponse delResp = destItem.Delete();
                    if(!delResp.IsSuccess)
                        return delResp;
                    newDestFolder = CopyThisItem(destFolder, null, destName);
                }
                else
                {
                    newDestFolder = destItem as Folder;
                    if(newDestFolder == null)
                        return new ConflictResponse();
                }
            }
            else
            {
                newDestFolder = CopyThisItem(destFolder, null, destName);
            }
            // move children
            MultistatusResponse mr = new MultistatusResponse();
            foreach(IHierarchyItem child in Children)
            {
                HierarchyItem dbchild = (HierarchyItem) child;
                WebDAVResponse resp = dbchild.MoveTo(newDestFolder, child.Name);
                if(!resp.IsSuccess)
                    mr.AddResponses(new ItemResponse(dbchild.Path, resp));
                else
                {
                    MultistatusResponse mrchild = resp as MultistatusResponse;
                    if(mrchild != null)
                        mr.AddResponses(mrchild.Responses);
                }
            }

            // delete this folder
            if(mr.Responses.Length > 0)
                return mr;
            else if(destItem == null)
            {
                DeleteThisItem();
                return new CreatedResponse();
            }
            else
            {
                DeleteThisItem();
                return new NoContentResponse();
            }
        }

		public override WebDAVResponse Delete()
		{
			Folder parent = Parent as Folder;
			if(parent == null)
				return new ConflictResponse();
			if(!parent.ClientHasToken)
				return new LockedResponse();

			if(!ClientHasToken)
				return new LockedResponse();

			MultistatusResponse mr = new MultistatusResponse();

			foreach(IHierarchyItem child in Children)
			{
                HierarchyItem dbchild = (HierarchyItem) child;
				WebDAVResponse resp = dbchild.Delete();
				if(!resp.IsSuccess)
					mr.AddResponses(new ItemResponse(dbchild.Path, resp));
				else
				{
					MultistatusResponse mrchild = resp as MultistatusResponse;
					if(mrchild != null)
						mr.AddResponses(mrchild.Responses);
				}
			}

			if(mr.Responses.Length > 0)
				return mr;
			else
			{
				DeleteThisItem();
				return new NoContentResponse();
			}
        }

        #endregion // HierarchyItem Members Override

        #region Non-public Members

        internal HierarchyItem FindChild(string childName)
        {
            HierarchyItem hierarchyItem = null;

            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();

                SqlCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT ID, ItemType, Name, Created, Modified"
                                  + " FROM Repository"
                                  + " WHERE Parent = @Parent"
                                  + " AND Name = @Name";

                cmd.Parameters.Add("@Parent", SqlDbType.Int).Value = ID;
                cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = childName;

                using (SqlDataReader reader = cmd.ExecuteReader())
                    if (reader.Read())
                    {
                        int itemID = reader.GetInt32(reader.GetOrdinal("ID"));
                        ItemType itemType = (ItemType) reader.GetByte(reader.GetOrdinal("ItemType"));
                        string itemName = reader.GetString(reader.GetOrdinal("Name"));
                        DateTime itemCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                        DateTime itemModified = reader.GetDateTime(reader.GetOrdinal("Modified"));

                        switch (itemType)
                        {
                            case ItemType.Resource:
                                hierarchyItem =
                                    new Resource(itemID, itemName, path + itemName, itemCreated, itemModified, engine);
                                break;
                            case ItemType.Folder:
                                hierarchyItem =
                                    new Folder(itemID, itemName, path + itemName + "/", itemCreated, itemModified,
                                               engine);
                                break;
                            case ItemType.LockNull:
                                hierarchyItem =
                                    new LockNull(itemID, itemName, path + itemName, itemCreated, itemModified, engine);
                                break;
                        }
                    }
            }

            return hierarchyItem;
        }

        internal void RemoveTree()
        {
            foreach(IHierarchyItem child in Children)
            {
                Folder childFolder = child as Folder;
                if(childFolder != null)
                    childFolder.RemoveTree();
                else
                    ((HierarchyItem)child).DeleteThisItem();
            }
            DeleteThisItem();
        }

        internal bool ClientHasTokenForTree()
        {
            if(!ClientHasToken)
                return false;
            foreach(IHierarchyItem child in Children)
            {
                Folder childFolder = child as Folder;
                if(childFolder != null)
                {
                    if(!childFolder.ClientHasTokenForTree())
                        return false;
                }
                else
                {
                    if (!((HierarchyItem) child).ClientHasToken)
                        return false;
                }
            }
            return true;
        }

        private HierarchyItem CreateChild(SqlTransaction trans, string childName, ItemType itemType)
        {
            SqlCommand cmd = trans.Connection.CreateCommand();
            cmd.Transaction = trans;
            cmd.CommandText = "INSERT INTO Repository"
                + " (Name, Created, Modified, Parent, ItemType)"
                + " VALUES(@Name, GETUTCDATE(), GETUTCDATE(), @Parent, @ItemType);"
                + "SELECT @Identity = @@IDENTITY";

            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = childName;
            cmd.Parameters.Add("@Parent", SqlDbType.Int).Value = ID;
            cmd.Parameters.Add("@ItemType", SqlDbType.TinyInt).Value = itemType;
            SqlParameter idParm = cmd.Parameters.Add("@Identity", SqlDbType.Int);
            idParm.Direction = ParameterDirection.Output;

            cmd.ExecuteNonQuery();

            int newID = (int)idParm.Value;

            UpdateModified(trans);

            switch (itemType)
            {
                case ItemType.Resource:
                    return new Resource(newID, childName, path + childName, DateTime.UtcNow, DateTime.UtcNow, engine);
                case ItemType.Folder:
                    // do not need to return created folder
                    return null;
                case ItemType.LockNull:
                    return new LockNull(newID, childName, path + childName, DateTime.UtcNow, DateTime.UtcNow, engine);
                default:
                    return null;
            }
        }

        #endregion // Non-public Members
    }
}
