using System;
using System.Collections;
using System.Collections.Specialized;
using System.Configuration;
using System.Data;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;
using System.Collections.Generic;

namespace WebDAVServer.SqlStorage
{
    public abstract class HierarchyItem : IHierarchyItem, ILock
    {
        protected int id;
        protected string name;
        protected string path;
        protected DateTime created;
        protected DateTime modified;
        protected string connStr = ConfigurationManager.AppSettings["SqlConnectionString"];
        protected readonly WDEngine engine;

        public HierarchyItem(int id, string name, string path, DateTime created, DateTime modified, WDEngine engine)
        {
            this.id = id;
            this.name = name;
            this.path = path;
            this.created = created;
            this.modified = modified;
            this.engine = engine;
        }

		#region IHierarchyItem Members

		public string Name
		{
			get { return name; }
		}

		public DateTime Created
		{
			get { return created; }
		}

		public DateTime Modified
		{
			get { return modified; }
		}

		public IFolder Parent
		{
			get
			{
				IFolder parent = null;

                using (SqlConnection conn = new SqlConnection(connStr))
                {
                    conn.Open();

                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SELECT Parent FROM Repository WHERE ID = @ID";
                    cmd.Parameters.Add("@ID", SqlDbType.Int).Value = ID;
                    int parentID = (int) cmd.ExecuteScalar();

                    if (parentID > 0)
                    {
                        cmd.CommandText = "SELECT Name, Created, Modified"
                                          + " FROM Repository"
                                          + " WHERE ID = @ID";

                        cmd.Parameters["@ID"].Value = parentID;

                        using (SqlDataReader reader = cmd.ExecuteReader())
                            if (reader.Read())
                            {
                                string parentName = reader.GetString(reader.GetOrdinal("Name"));
                                DateTime parentCreated = reader.GetDateTime(reader.GetOrdinal("Created"));
                                DateTime parentModified = reader.GetDateTime(reader.GetOrdinal("Modified"));

                                int ln = name.Length;
                                int lp = path.Length;
                                parent =
                                    new Folder(parentID, parentName, path.Remove(lp - ln, ln), parentCreated,
                                               parentModified, engine);
                            }
                    }
                }

			    return parent;
			}
		}

        public virtual string Path
        {
            get
            {
                return this.path;
            }
        }

		public abstract WebDAVResponse CopyTo(IFolder folder, string destName, bool deep);

		public abstract WebDAVResponse MoveTo(IFolder folder, string destName);

		public abstract WebDAVResponse Delete();

		public WebDAVResponse GetProperties(ref Property[] props)
		{
            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();

                SqlCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT Name, Namespace, PropVal"
                                  + " FROM Properties"
                                  + " WHERE ItemID = @ItemID";

                cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = ID;

                using (SqlDataReader reader = cmd.ExecuteReader())
                {
                    if (props == null) // get all properties
                    {
                        List<Property> l = new List<Property>();
                        while (reader.Read())
                        {
                            Property p = new Property();
                            p.Name = reader.GetString(reader.GetOrdinal("Name"));
                            p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                            p.Value = reader.GetString(reader.GetOrdinal("PropVal"));
                            l.Add(p);
                        }
                        props = l.ToArray();
                    }
                    else // get selected properties
                    {
                        Property p = new Property();
                        while (reader.Read())
                        {
                            p.Name = reader.GetString(reader.GetOrdinal("Name"));
                            p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                            for (int i = 0; i < props.Length; i++)
                                if (p.Name == props[i].Name && p.Namespace == props[i].Namespace)
                                {
                                    p.Value = reader.GetString(reader.GetOrdinal("PropVal"));
                                    props[i] = p;
                                    break;
                                }
                        }
                    }
                }
            }

		    return new OkResponse();
		}

		public virtual WebDAVResponse UpdateProperties(Property[] setProps, Property[] delProps)
		{
			if(!ClientHasToken)
				return new LockedResponse();
            
			MultipropResponse resp = new MultipropResponse();

		    using (SqlConnection conn = new SqlConnection(connStr))
		    {
		        SqlTransaction trans = null;
                try
                {
                    conn.Open();
                    trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted);

                    if (setProps != null)
                        foreach (Property p in setProps)
                        {
                            SetProperty(p, trans); // create or update property
                            resp.AddResponses(new PropResponse(p, new OkResponse()));
                        }

                    if (delProps != null)
                        foreach (Property p in delProps)
                        {
                            RemoveProperty(p.Name, p.Namespace, trans);
                            resp.AddResponses(new PropResponse(p, new OkResponse()));
                        }

                    UpdateModified(trans);

                    trans.Commit();
                }
                catch
                {
                    if (trans != null) trans.Rollback();
                    return new ServerErrorResponse();
                }
		    }

		    return resp;
		}

        public WebDAVResponse GetPropertyNames(ref Property[] props)
        {
            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();

                SqlCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT Name, Namespace"
                                  + " FROM Properties"
                                  + " WHERE ItemID = @ItemID";

                cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = ID;

                List<Property> l = new List<Property>();
                using (SqlDataReader reader = cmd.ExecuteReader())
                    while (reader.Read())
                    {
                        Property p = new Property();
                        p.Name = reader.GetString(reader.GetOrdinal("Name"));
                        p.Namespace = reader.GetString(reader.GetOrdinal("Namespace"));
                        l.Add(p);
                    }
                props = l.ToArray();
            }

            return new OkResponse();
        }

        #endregion // IHierarchyItem Members

        #region ILock Members

        public WebDAVResponse Lock(ref LockInfo lockInfo)
		{
			if(ItemHasLock(lockInfo.Shared))
				return new LockedResponse();

			if(lockInfo.Deep)
			{ // check if no items are locked in this subtree
				MultistatusResponse mr = FindLocksDown(this, lockInfo.Shared);
				if(mr.Responses.Length > 0)
					return mr;
			}

            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();
                SqlCommand cmd = conn.CreateCommand();
                cmd.CommandText = "INSERT INTO Locks (ItemID,Token,Shared,Deep,Expires,Owner)"
                                  + " VALUES(@ItemID, @Token, @Shared, @Deep, @Expires, @Owner)";

                lockInfo.Token = Guid.NewGuid().ToString();

                cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = ID;
                cmd.Parameters.Add("@Token", SqlDbType.Char).Value = lockInfo.Token;
                cmd.Parameters.Add("@Shared", SqlDbType.Bit).Value = lockInfo.Shared;
                cmd.Parameters.Add("@Deep", SqlDbType.Bit).Value = lockInfo.Deep;
                cmd.Parameters.Add("@Expires", SqlDbType.DateTime);
                if (lockInfo.Timeout >= 0)
                {
                    cmd.Parameters["@Expires"].Value = DateTime.Now.AddSeconds(lockInfo.Timeout);
                }
                else
                {
                    cmd.Parameters["@Expires"].Value = SqlDateTime.Null; // no expiration
                }
                cmd.Parameters.Add("@Owner", SqlDbType.NVarChar).Value = lockInfo.Owner;

                cmd.ExecuteNonQuery();
            }

            return new OkResponse();
		}
    
		public WebDAVResponse RefreshLock(ref LockInfo lockInfo)
		{
			LockInfo[] locks = ActiveLocks;

			int lockIndex;
			for(lockIndex = 0; lockIndex < locks.Length; lockIndex++)
				if(lockInfo.Token == locks[lockIndex].Token) break;

			if(lockIndex == locks.Length)
				return new PreconditionFailedResponse();

            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();

                SqlCommand cmd = conn.CreateCommand();
                cmd.CommandText = "UPDATE Locks SET Expires = @Expires WHERE Token = @Token";

                cmd.Parameters.Add("@Expires", SqlDbType.DateTime);
                if (lockInfo.Timeout >= 0)
                {
                    cmd.Parameters["@Expires"].Value = DateTime.Now.AddSeconds(lockInfo.Timeout);
                }
                else
                {
                    cmd.Parameters["@Expires"].Value = SqlBoolean.Null;
                }
                cmd.Parameters.Add("@Token", SqlDbType.Char).Value = lockInfo.Token;

                cmd.ExecuteNonQuery();
            }

		    // return info about lock to WebDAV engine
			lockInfo.Shared = locks[lockIndex].Shared;
			lockInfo.Deep = locks[lockIndex].Deep;
			lockInfo.Owner = locks[lockIndex].Owner;

			return new OkResponse();
		}

		public WebDAVResponse Unlock(string lockToken)
		{
			LockInfo[] locks = ActiveLocks;
			int i;
            
			for(i=0;i<locks.Length;i++)
				if(locks[i].Token == lockToken) break;
            
			if(i == locks.Length)
				return new PreconditionFailedResponse();

			if(this is ILockNull)
			{ // delete lock-null item
				DeleteThisItem();
			}
			else // remove lock from existing item
                using (SqlConnection conn = new SqlConnection(connStr))
                {
                    conn.Open();

                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "DELETE FROM Locks WHERE Token = @Token";
                    cmd.Parameters.Add("@Token", SqlDbType.NVarChar).Value = lockToken;
                    cmd.ExecuteNonQuery();
                }

		    return new NoContentResponse();
		}

		public LockInfo[] ActiveLocks
		{
			get
			{
				int itemID = ID;
				ArrayList l = new ArrayList();

                using (SqlConnection conn = new SqlConnection(connStr))
                {
                    conn.Open();

                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SELECT Parent FROM Repository WHERE ID = @ID";
                    cmd.Parameters.Add("@ID", SqlDbType.Int);
                    cmd.Prepare();

                    l.AddRange(GetLocks(ID, false)); // get all locks
                    while (true)
                    {
                        cmd.Parameters["@ID"].Value = itemID;
                        itemID = (int) cmd.ExecuteScalar();
                        if (itemID <= 0) break;
                        l.AddRange(GetLocks(itemID, true)); // get only deep locks
                    }
                }

			    return (LockInfo[])l.ToArray(typeof(LockInfo));
			}
        }

        #endregion // ILock Members

        #region user properties methods

        private void SetProperty(Property prop, SqlTransaction trans)
        {
            SqlCommand cmd = trans.Connection.CreateCommand();
            cmd.Transaction = trans;
            
            cmd.CommandText = "SELECT Count(*) FROM Properties"
                +" WHERE ItemID = @ItemID"
                +" AND Name = @Name"
                +" AND Namespace = @Namespace";
            
            cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = ID;
            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = prop.Name;
            cmd.Parameters.Add("@Namespace", SqlDbType.NVarChar).Value = prop.Namespace;

            int count = (int) cmd.ExecuteScalar();
            
            if(count == 0) // insert
            {
                cmd.CommandText = "INSERT INTO Properties"
                    +" (ItemID, Name, Namespace, PropVal)"
                    +" VALUES(@ItemID, @Name, @Namespace, @PropVal)";
                
                cmd.Parameters.Add("@PropVal", SqlDbType.NVarChar).Value = prop.Value;
            }
            else // update
            {
                cmd.CommandText = "UPDATE Properties"
                    +" SET PropVal = @PropVal"
                    +" WHERE ItemID = @ItemID"
                    +" AND Name = @Name"
                    +" AND Namespace = @Namespace";
                
                cmd.Parameters.Add("@PropVal", SqlDbType.NVarChar).Value = prop.Value;
            }

            cmd.ExecuteNonQuery();
        }

        private void RemoveProperty(string propertyName, string ns, SqlTransaction trans)
        {
            SqlCommand cmd = trans.Connection.CreateCommand();
            cmd.Transaction = trans;
            cmd.CommandText = "DELETE FROM Properties"
                +" WHERE ItemID = @ItemID"
                +" AND Name = @Name"
                +" AND Namespace = @Namespace";
            
            cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = ID;
            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = propertyName;
            cmd.Parameters.Add("@Namespace", SqlDbType.NVarChar).Value = ns;

            cmd.ExecuteNonQuery();
        }
        #endregion

        #region copy methods

        internal Folder CopyThisItem(Folder destFolder, HierarchyItem destItem, string destName)
        {
            // returns created folder, if any, otherwise null

            Folder createdFolder = null;

            using (SqlConnection conn = new SqlConnection(connStr))
            {
                SqlTransaction trans = null;
                conn.Open();

                try
                {
                    trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted);
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.Transaction = trans;

                    int destID;
                    if(destItem == null) // insert new item
                    {
                        // copy item
                        cmd.CommandText = "INSERT INTO Repository"
                                          + " (Name, Created, Modified, Parent, ItemType, ContentType, Content)"
                                          + " SELECT @Name, GETUTCDATE(), GETUTCDATE(), @Parent, ItemType, ContentType, Content"
                                          + " FROM Repository"
                                          + " WHERE ID = @ID;"
                                          + "SELECT @Identity = @@IDENTITY";
                    
                        cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = destName;
                        cmd.Parameters.Add("@Parent", SqlDbType.Int).Value = destFolder.ID;
                        cmd.Parameters.Add("@ID", SqlDbType.Int).Value = ID;
                        SqlParameter idParm = cmd.Parameters.Add("@Identity", SqlDbType.Int);
                        idParm.Direction = ParameterDirection.Output;
                    
                        cmd.ExecuteNonQuery();
                    
                        destID = (int)idParm.Value;

                        destFolder.UpdateModified(trans);
                
                        if(this is IFolder)
                        {
                            createdFolder = new Folder(destID,destName,destFolder.path+destName+"/",
                                                       DateTime.UtcNow,DateTime.UtcNow, engine);
                        }
                    }
                    else // update existing destination
                    {
                        destID = destItem.ID;

                        cmd.CommandText = "UPDATE Repository SET"
                                          + " Modified = GETUTCDATE()"
                                          + ", ItemType = src.ItemType"
                                          + ", ContentType = src.ContentType"
                                          + ", Content = src.Content"
                                          + " FROM (SELECT * FROM Repository WHERE ID=@SrcID) src"
                                          + " WHERE Repository.ID=@DestID";

                        cmd.Parameters.Add("@SrcID", SqlDbType.Int).Value = ID;
                        cmd.Parameters.Add("@DestID", SqlDbType.Int).Value = destID;

                        cmd.ExecuteNonQuery();

                        // remove old properties from the destination
                        cmd.CommandText = "DELETE FROM Properties WHERE ItemID = @ItemID";
                    
                        cmd.Parameters.Clear();
                        cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = destID;
                    
                        cmd.ExecuteNonQuery();
                    }

                    // copy properties
                    cmd.CommandText = "INSERT INTO Properties"
                                      +" (ItemID, Name, Namespace, PropVal)"
                                      +" SELECT @DestID, Name, Namespace, PropVal"
                                      +" FROM Properties"
                                      +" WHERE ItemID = @SrcID";

                    cmd.Parameters.Clear();
                    cmd.Parameters.Add("@SrcID", SqlDbType.Int).Value = ID;
                    cmd.Parameters.Add("@DestID", SqlDbType.Int).Value = destID;

                    cmd.ExecuteNonQuery();

                    trans.Commit();
                }
                catch(Exception)
                {
                    if(trans != null) trans.Rollback();
                    throw;
                }
            }

            return createdFolder;
        }

        #endregion

		#region move methods

		internal void MoveThisItem(Folder destFolder, string destName, Folder parent)
		{
		    using (SqlConnection conn = new SqlConnection(connStr))
		    {
		        SqlTransaction trans = null;
		        conn.Open();

		        try
		        {
		            trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted);
		            SqlCommand cmd = conn.CreateCommand();
		            cmd.Transaction = trans;
            
		            // remove lock from item
		            cmd.CommandText = "DELETE FROM Locks WHERE ItemID = @ItemID";
		            cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = ID;
		            cmd.ExecuteNonQuery();

		            // move item
		            cmd.CommandText = "UPDATE Repository SET"
		                              +" Name = @Name"
		                              +", Parent = @Parent"
		                              +" WHERE ID = @ItemID";
                
		            cmd.Parameters.Add("@Name", SqlDbType.NVarChar).Value = destName;
		            cmd.Parameters.Add("@Parent", SqlDbType.Int).Value = destFolder.ID;
                
		            cmd.ExecuteNonQuery();
                
		            parent.UpdateModified(trans);
		            destFolder.UpdateModified(trans);

		            trans.Commit();
            
		            name = destName;
		        }
		        catch(Exception)
		        {
		            if(trans != null) trans.Rollback();
		            throw;
		        }
		    }
		}

		#endregion

        #region delete methods

        internal void DeleteThisItem()
        {
            Folder parentFolder = Parent as Folder;

            using (SqlConnection conn = new SqlConnection(connStr))
            {
                SqlTransaction trans = null;
                conn.Open();

                try
                {
                    trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted);
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.Transaction = trans;

                    // delete item locks
                    cmd.CommandText = "DELETE FROM Locks WHERE ItemID = @ItemID";
                    cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = ID;
                    cmd.ExecuteNonQuery();
                
                    // delete item properties
                    cmd.CommandText = "DELETE FROM Properties WHERE ItemID = @ItemID";
                    cmd.ExecuteNonQuery();
                
                    // delete item
                    cmd.CommandText = "DELETE FROM Repository WHERE ID = @ItemID";
                    cmd.ExecuteNonQuery();
                
                    if(parentFolder != null) parentFolder.UpdateModified(trans);

                    trans.Commit();
                }
                catch(Exception)
                {
                    if(trans != null) trans.Rollback();
                    throw;
                }
            }
        }

        #endregion

		#region lock methods

		private ArrayList GetLocks(int itemID, bool onlyDeep)
		{
			ArrayList l = new ArrayList(); // list of active locks

            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();

                SqlCommand cmd = conn.CreateCommand();
                if (onlyDeep)
                {
                    cmd.CommandText = "SELECT Token, Shared, Deep, Expires, Owner"
                                      + " FROM Locks"
                                      + " WHERE ItemID = @ItemID"
                                      + " AND Deep = @Deep";
                    cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = itemID;
                    cmd.Parameters.Add("@Deep", SqlDbType.Bit).Value = true;
                }
                else
                {
                    cmd.CommandText = "SELECT Token, Shared, Deep, Expires, Owner FROM Locks WHERE ItemID = @ItemID";
                    cmd.Parameters.Add("@ItemID", SqlDbType.Int).Value = itemID;
                }
                using (SqlDataReader reader = cmd.ExecuteReader())
                    while (reader.Read())
                    {
                        LockInfo li = new LockInfo();
                        li.Token = reader.GetString(reader.GetOrdinal("Token"));
                        li.Shared = reader.GetBoolean(reader.GetOrdinal("Shared"));
                        li.Deep = reader.GetBoolean(reader.GetOrdinal("Deep"));
                        if (reader.IsDBNull(reader.GetOrdinal("Expires")))
                        {
                            li.Timeout = -1;
                        }
                        else
                        {
                            DateTime expires = reader.GetDateTime(reader.GetOrdinal("Expires"));
                            if (expires <= DateTime.Now)
                            {
                                li.Timeout = 0;
                            }
                            else
                                li.Timeout = (expires.Ticks - DateTime.Now.Ticks)/10000000;
                        }
                        li.Owner = reader.GetString(reader.GetOrdinal("Owner"));

                        if (li.Timeout != 0)
                            l.Add(li);
                    }
            }

		    return l;
		}

		internal bool ClientHasToken
		{
			get
			{
				LockInfo[] itemLocks = ActiveLocks;
				if(itemLocks.Length == 0) 
					return true;
				//WDRequest request = new WDRequest();
				StringCollection clientLockTokens =  engine.Request.ClientLockTokens;
				for(int i = 0; i<clientLockTokens.Count; i++)
					for(int j = 0; j<itemLocks.Length; j++)
						if(clientLockTokens[i] == itemLocks[j].Token) 
							return true;
				return false;
			}
		}

		protected bool ItemHasLock(bool skipShared)
		{
			LockInfo[] locks = ActiveLocks;
			if(locks.Length == 0)
				return false;
			return !(skipShared && locks[0].Shared);
		}

		protected static MultistatusResponse FindLocksDown(IHierarchyItem root, bool skipShared)
		{
			MultistatusResponse mr = new MultistatusResponse();
			IFolder folder = root as IFolder;
			if(folder != null)
				foreach(IHierarchyItem child in folder.Children)
				{
                    HierarchyItem dbchild = (HierarchyItem)child;
					if(dbchild.ItemHasLock(skipShared))
						mr.AddResponses(new ItemResponse(dbchild.Path, new LockedResponse()));
					MultistatusResponse mrchild = FindLocksDown(child, skipShared);
					mr.AddResponses(mrchild.Responses);
				}
			return mr;
		}

		#endregion

        #region Non-public Members

        internal void UpdateModified(SqlTransaction trans)
		{
			SqlCommand cmd = trans.Connection.CreateCommand();
			cmd.Transaction = trans;
            
			cmd.CommandText = "UPDATE Repository SET Modified = GETUTCDATE() WHERE ID = @ID";
			cmd.Parameters.Add("@ID", SqlDbType.Int).Value = ID;
            
			cmd.ExecuteNonQuery();
		}

		protected static bool DelWhenReplace
		{
			get
			{
                switch (ConfigurationManager.AppSettings["DeleteWhenReplace"])
				{
					case "true":
						return true;
					case "false":
						return false;
					default:
						return true;
				}
			}
		}

		internal int ID
		{
			get { return id; }
        }

        #endregion // Non-public Members
    }
}
