using System;
using System.Data;
using System.Data.SqlClient;
using System.IO;
using System.Web;
using ITHit.WebDAV.Server;
using ITHit.WebDAV.Server.Response;

namespace WebDAVServer.SqlStorage
{
    public class Resource : HierarchyItem, IResource
    {
        public Resource(int ID, string name, string path, DateTime created, DateTime modified, WDEngine engine)
            :base(ID, name, path, created, modified, engine)
        {
        }
        
        #region IResource Members

        public string ContentType
        {
            get
            {
                string contentType = "";

                using (SqlConnection conn = new SqlConnection(connStr))
                {
                    conn.Open();
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SELECT ContentType FROM Repository WHERE ID = @ID";
                    cmd.Parameters.Add("@ID", SqlDbType.Int).Value = ID;
                    using (SqlDataReader reader = cmd.ExecuteReader())
                    {
                        reader.Read();
                        if (!reader.IsDBNull(reader.GetOrdinal("ContentType")))
                            contentType = reader.GetString(reader.GetOrdinal("ContentType"));
                    }
                }

                if (contentType == "")
                    contentType =
                        MimeType.GetMimeType(System.IO.Path.GetExtension(this.name))
                        ?? "application/octet-stream";
                return contentType;
            }
        }

        public long ContentLength
        {
            get
            {
                long length = 0;

                using (SqlConnection conn = new SqlConnection(connStr))
                {
                    conn.Open();
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.CommandText = "SELECT Content FROM Repository WHERE ID = @ID";
                    cmd.Parameters.Add("@ID", SqlDbType.Int).Value = ID;
                    using (SqlDataReader reader = cmd.ExecuteReader(CommandBehavior.SequentialAccess))
                    {
                        reader.Read();
                        if (!reader.IsDBNull(reader.GetOrdinal("Content")))
                            length = reader.GetBytes(reader.GetOrdinal("Content"), 0, null, 0, 0);
                    }
                }

                return length;
            }
        }

        public WebDAVResponse WriteToStream(Stream output, long byteStart, long count)
        {
            HttpContext.Current.Server.ScriptTimeout = 2400; // timeout in seconds.

            using (SqlConnection conn = new SqlConnection(connStr))
            {
                conn.Open();
                SqlCommand cmd = conn.CreateCommand();
                cmd.CommandText = "SELECT Content FROM Repository WHERE ID = @ID";
                cmd.Parameters.Add("@ID", SqlDbType.Int).Value = ID;

                using (SqlDataReader reader = cmd.ExecuteReader(CommandBehavior.SequentialAccess))
                {
                    reader.Read();

                    long bufSize = 1048576; // 1Mb
                    byte[] buf = new byte[bufSize];
                    long retval;
                    long startIndex = byteStart;

                    while ((retval = reader.GetBytes(reader.GetOrdinal("Content"),
                                                     startIndex, buf, 0, (int) (count > bufSize ? bufSize : count))) > 0)
                    {
                        try
                        {
                            output.Write(buf, 0, (int) retval);
                        }
                        catch (HttpException)
                        {
                            // The remote host closed the connection
                            return new OkResponse();
                        }

                        startIndex += retval;
                        count -= retval;
                    }
                }
            }

            return new OkResponse();
        }

        public WebDAVResponse SaveFromStream(Stream content, string contentType)
        {
            if (!ClientHasToken)
                return new LockedResponse();

            WebDAVResponse resp;

            using (SqlConnection conn = new SqlConnection(connStr))
            {
                SqlTransaction trans = null;
                conn.Open();
            
                try
                {
                    trans = conn.BeginTransaction(IsolationLevel.ReadUncommitted);
                    SqlCommand cmd = conn.CreateCommand();
                    cmd.Transaction = trans;
                    long read = WriteContent(trans, content, contentType);
                    if (read > 0)
                    {
                        resp = new OkResponse();
                    }
                    else
                    {
                        resp = new NoContentResponse();
                    }
                    trans.Commit();
                }
                catch
                {
                    if(trans != null) trans.Rollback();
                    resp = new ServerErrorResponse();
                }
            }

            return resp;
        }

        #endregion // IResource Members

        #region HierarchyItem Methods Override

        public override WebDAVResponse CopyTo(IFolder folder, string destName, bool deep)
        {
            Folder destFolder = folder as Folder;
            if(destFolder == null)
                return new ConflictResponse();
            if(!destFolder.ClientHasToken)
                return new LockedResponse();

            HierarchyItem destItem = destFolder.FindChild(destName);
            if(destItem != null)
            {
                if(DelWhenReplace && destItem is IFolder)
                {
                    Folder destItemF = (Folder) destItem;
                    if(!destItemF.ClientHasTokenForTree())
                        return new LockedResponse();
                    destItemF.RemoveTree();
                    CopyThisItem(destFolder,null,destName);
                    return new NoContentResponse();
                }
                else if(destItem is IResource)
                {
                    if(destItem.ClientHasToken)
                    {
                        CopyThisItem(null,destItem,destName);
                        return new NoContentResponse();
                    }
                    else // destination is locked
                        return new LockedResponse();
                }
                else
                    return new ConflictResponse();
            }
            else
            {
                CopyThisItem(destFolder,null,destName);
                return new CreatedResponse();
            }
        }

        public override WebDAVResponse Delete()
        {
            Folder parent = Parent as Folder;
            if(parent == null)
                return new ConflictResponse();
            if(!parent.ClientHasToken)
                return new LockedResponse();

            if(!ClientHasToken)
                return new LockedResponse();
            
            DeleteThisItem();
            return new NoContentResponse();
        }
        
        public override WebDAVResponse MoveTo(IFolder folder, string destName)
        {
            Folder destFolder = folder as Folder;
            if(destFolder == null)
                return new ConflictResponse();
            
            Folder parent = Parent as Folder;
            if(parent == null)
                return new ConflictResponse();
            
            if(!ClientHasToken || !destFolder.ClientHasToken || !parent.ClientHasToken)
                return new LockedResponse();
            
            HierarchyItem destItem = destFolder.FindChild(destName);
            if(destItem != null)
            {
                if(DelWhenReplace && destItem is IFolder)
                {
                    Folder destItemF = (Folder) destItem;
                    if(!destItemF.ClientHasTokenForTree())
                        return new LockedResponse();
                    destItemF.RemoveTree();
                    MoveThisItem(destFolder, destName, parent);
                    return new NoContentResponse();
                }
                else if(destItem is IResource)
                {
                    WebDAVResponse delResp = destItem.Delete();
                    if(!delResp.IsSuccess)
                        return delResp;
                    MoveThisItem(destFolder, destName, parent);
                    return new NoContentResponse();
                }
                else
                    return new ConflictResponse();
            }
            else
            {
                MoveThisItem(destFolder, destName, parent);
                return new CreatedResponse();
            }
        }

        #endregion // HierarchyItem Methods Override
        
        internal long WriteContent(SqlTransaction trans, Stream content, string contentType)
        {
            HttpContext.Current.Server.ScriptTimeout = 2400;

            SqlCommand cmd = trans.Connection.CreateCommand();
            cmd.Transaction = trans;

            long read = 0;
            //Length is unknown for POST
            //if (content.Length > 0)
            {
                int bufSize = 1048576; // 1Mb

                cmd.CommandText = "UPDATE Repository SET"
                    + " Modified = GETUTCDATE(),"
                    + " Content = NULL,"
                    + " ContentType = @ContentType"
                    + " WHERE ID = @ID;"
                    + "SELECT @Pointer = TEXTPTR(Content) FROM Repository WHERE ID = @ID";

                if (contentType != null)
                    cmd.Parameters.Add("@ContentType", SqlDbType.NVarChar).Value = contentType;
                else
                    cmd.Parameters.Add("@ContentType", SqlDbType.NVarChar).Value = DBNull.Value;
                cmd.Parameters.Add("@ID", SqlDbType.Int).Value = ID;

                SqlParameter ptrParm = cmd.Parameters.Add("@Pointer", SqlDbType.Binary, 16);
                ptrParm.Direction = ParameterDirection.Output;

                cmd.ExecuteNonQuery();

                cmd.CommandText = "UPDATETEXT Repository.Content @Pointer @Offset NULL @Bytes";

                cmd.Parameters.Clear();
                cmd.Parameters.Add("@Pointer", SqlDbType.Binary, 16).Value = ptrParm.Value;
                SqlParameter offsetParm = cmd.Parameters.Add("@Offset", SqlDbType.Int);
                SqlParameter bytesParm = cmd.Parameters.Add("@Bytes", SqlDbType.Image, bufSize);

                BinaryReader br = new BinaryReader(content);

                byte[] buf = br.ReadBytes(bufSize);
                int offset_ctr = 0;

                while (buf.Length > 0)
                {
                    read += buf.Length;
                    bytesParm.Value = buf;
                    offsetParm.Value = offset_ctr;
                    cmd.ExecuteNonQuery();
                    offset_ctr += buf.Length;
                    buf = br.ReadBytes(bufSize);
                }

                br.Close();
            }
            /*else
            {
                cmd.CommandText = "UPDATE Repository SET"
                    + " Modified = GETUTCDATE(),"
                    + " Content = NULL,"
                    + " ContentType = @ContentType"
                    + " WHERE ID = @ID";
                    
                if (contentType != null)
                    cmd.Parameters.Add("@ContentType", SqlDbType.NVarChar).Value = contentType;
                else
                    cmd.Parameters.Add("@ContentType", SqlDbType.NVarChar).Value = DBNull.Value;
                cmd.Parameters.Add("@ID", SqlDbType.Int).Value = ID;

                cmd.ExecuteNonQuery();
            }*/

            return read;
        }

    }
}
